'''This module is about disposing graphical elements in the user interface'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.px import Px

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# A "layout" defines how to dispose, in the user interface, parts of some
# logical element. This element, named a "layouted object", can be a field or a
# page. 

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Cell:
    '''Represents a cell in a row within a layout (see appy.ui.layout.Layout)'''

    def __init__(self, content, align, isHeader=False):
        self.align = align
        self.width = None
        self.content = None
        self.colspan = 1
        if isHeader:
            self.width = content
        else:
            self.content = [] # The list of widgets to render in the cell
            self.decodeContent(content)

    def __repr__(self):
        return '<Cell %s>' % self.content

    def decodeContent(self, content):
        digits = '' # We collect the digits that will give the colspan
        for char in content:
            if char.isdigit():
                digits += char
            else: # It is a letter corresponding to a macro
                self.content.append(Layout.pxs.get(char, char))
        # Manage the colspan
        if digits:
            self.colspan = int(digits)

    def renderContent(self, value, layout, layoutTarget):
        '''Renders p_value (one element among self.content) for a given
           p_layoutTarget (a field or object) on some p_layout.'''
        # Do not render "r" if we must render a not-required object
        if (value == 'pxRequired') and not layoutTarget.required: return
        # The name of the PX is the layout or p_value
        name = layout if value == 'f' else value
        r = layoutTarget
        if isinstance(name, tuple):
            for part in name:
                r = getattr(r, part)
        else:
            r = getattr(r, name)
        return r

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Row:
    '''Represents a a row within a layout (see appy.ui.layout.Layout)'''

    def __init__(self, content, valign, isHeader=False):
        self.valign = valign
        self.cells = []
        self.decodeCells(content, isHeader)
        # Compute the row length
        length = 0
        for cell in self.cells:
            length += cell.colspan
        self.length = length

    def __repr__(self):
        return '<Row %s (%d)>' % (str(self.cells), self.length)

    def decodeCells(self, content, isHeader):
        '''Decodes the given chunk of layout string p_content containing
           column-related information (if p_isHeader is True) or cell content
           (if p_isHeader is False) and produces a list of Cell instances.'''
        cellContent = ''
        for char in content:
            if char in Layout.cellDelimiters:
                align = Layout.cellDelimiters[char]
                self.cells.append(Cell(cellContent, align, isHeader))
                cellContent = ''
            else:
                cellContent += char
        # Manage the last cell if any
        if cellContent:
            self.cells.append(Cell(cellContent, 'left', isHeader))

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Layout:
    '''Represents a logical table defining how to dispose graphical parts of
       some UI element.'''

    # --------------------------------------------------------------------------
    # Defining parts of a layouted object and how they will be rendered is done
    # via a "layout string". Such a string is made of:
    # - letters, corresponding to object parts;
    # - symbols, allowing to define column or row breaks, alignment, etc.

    # Letters for a page -------------------------------------------------------
    #  e - The object h*e*ader, containing its title, breadcrumb, siblings and
    #      navigation.
    #  b - The range of *b*uttons (pages and phases, save, delete, actions and
    #      workflow transitions.
    #  w - The fields (= *w*idgets) to render on some page for some object

    # Letters for a field ------------------------------------------------------
    #  l - "label"        The field label
    #  d - "description"  The field description
    #  h - "help"         Help for the field (typically rendered as an icon,
    #                     with a tooltip displaying online help
    #  v - "validation"   The icon that is shown when a validation error occurs
    #                     (typically only used on "edit" layouts)
    #  r - "required"     The asterisc meaning that the field is required
    #                     (if relevant; typically only used on "edit" layouts)
    #  f - "field"        The field value, or input for entering a value.
    #  c - "changes"      The button for displaying changes to a field

    # This dict defines the PXs to use to render every part as defined by the
    # hereabove letters.
    pxs = {
      # Page-related elements
      'e': 'pxHeader', 'w': 'pxFields', 'b': 'pxControls',
      # Field-related elements
      'l': 'pxLabel', 'd': 'pxDescription', 'h': 'pxHelp', 'v': 'pxValidation',
      'r': 'pxRequired', 'c': 'pxChanges'
    }

    # Symbols to use within a layout string ------------------------------------
    # A row delimiter is to be used at the end of a row. The symbol used defines
    # alignment for the previously defined row.
    rowDelimiters =  {'-':'middle', '=':'top', '_':'bottom'}
    rowDelms = ''.join(rowDelimiters.keys())
    # A cell delimiter is to be used at the end of a cell. The symbol used
    # defines alignment for the previouly defined cell.
    cellDelimiters = {'|': 'center', ';': 'left', '!': 'right'}
    cellDelms = ''.join(cellDelimiters.keys())

    # Base layout attributes
    baseAttributes = ('style', 'css_class', 'cellpadding', 'cellspacing',
                      'width', 'align')

    # A layout can be derived from another one. The following dict defines the
    # letters to remove to define a simpler layout (view) from a more complex
    # layout (edit).
    derivedRepls = {'view': 'hrvd'}

    # PX rendering this layout instance, known in the context as "table".
    # Indeed, the variable named "layout" is already used to represent the
    # current layout *type*. If the layouted object is a page, the "layout
    # target" (where to look for sub-PXs) will be the object whose page is
    # shown; if the layouted object is a field, the layout target will be this
    # field.

    # Warning: when the layout type is "cell", the cell width and alignment must
    # not be defined by the cell layout, but by the outer column layout.

    pxRender = Px('''
     <table var="layoutCss=table.css_class;
                 inTd=(layout == 'cell') and bool(column)|False"
       cellpadding=":table.cellpadding" cellspacing=":table.cellspacing"
       width=":'' if inTd else table.width"
       align=":'' if inTd else ui.Language.flip(table.align, dir)"
       class=":('%s %s' % (tagCss, layoutCss)).strip() if tagCss else layoutCss"
       style=":table.style" id=":tagId" name=":tagName">
      <!-- The table header row -->
      <tr if="table.headerRow" valign=":table.headerRow.valign">
       <th for="cell in table.headerRow.cells" width=":cell.width"
           align=":ui.Language.flip(cell.align, dir)">
       </th>
      </tr>
      <!-- The table content -->
      <tr for="row in table.rows" valign=":row.valign">
       <td for="cell in row.cells" colspan=":cell.colspan"
           align=":ui.Language.flip(cell.align, dir)"
           class=":'' if loop.cell.last else 'cellGap'">
        <x for="c in cell.content">
         <x>::cell.renderContent(c, layout, layoutTarget)</x>
        </x>
       </td>
      </tr>
     </table>''')

    def __init__(self, layoutString=None, style=None, css_class='',
                 cellpadding=0, cellspacing=0, width='100%', align='left',
                 other=None, derivedType=None):
        if other:
            # We need to create a Table instance from another Table instance,
            # given in p_other. In this case, we ignore previous params.
            if derivedType is not None:
                # We will not simply mimic p_other. If p_derivedType is:
                # - "view", p_other is an "edit" layout, and we must create the
                #           corresponding "view" layout;
                # - "search", p_derivedFrom is a "view" layout.
                self.layoutString = Layout.derive(other.layoutString,
                                                  derivedType)
            else:
                self.layoutString = other.layoutString
            source = 'other.'
        else:
            source = ''
            self.layoutString = layoutString
        # Initialise simple params, either from the true params, either from
        # the p_other Table instance.
        for param in Layout.baseAttributes:
            setattr(self, param, eval('%s%s' % (source, param)))
        # The following attribute will store a special Row instance used for
        # defining column properties.
        self.headerRow = None
        # The content rows will be stored hereafter
        self.rows = []
        self.decodeRows(self.layoutString)

    @staticmethod
    def derive(layout, derivedType):
        '''Returns a layout derived from p_layout'''
        res = layout
        for letter in Layout.derivedRepls[derivedType]:
            res = res.replace(letter, '')
        # Strip the derived layout
        res = res.lstrip(Layout.rowDelms); res = res.lstrip(Layout.cellDelms)
        return res

    def addCssClasses(self, css_class):
        '''Adds a single or a group of p_css_class.'''
        if not self.css_class: self.css_class = css_class
        else:
            self.css_class += ' ' + css_class
            # Ensures that every class appears once
            self.css_class = ' '.join(set(self.css_class.split()))

    def isHeaderRow(self, rowContent):
        '''Determines if p_rowContent specified the table header row or a
           content row.'''
        # Find the first char that is a number or a letter
        for char in rowContent:
            if char not in Layout.cellDelimiters:
                return char.isdigit()
        return True

    def decodeRows(self, layoutString):
        '''Decodes the given p_layoutString and produces a list of Row
           instances.'''
        # Split the p_layoutString with the row delimiters
        rowContent = ''
        for char in layoutString:
            if char in Layout.rowDelimiters:
                valign = Layout.rowDelimiters[char]
                if self.isHeaderRow(rowContent):
                    if not self.headerRow:
                        self.headerRow = Row(rowContent, valign, isHeader=True)
                else:
                    self.rows.append(Row(rowContent, valign))
                rowContent = ''
            else:
                rowContent += char
        # Manage the last row if any
        if rowContent:
            self.rows.append(Row(rowContent, 'middle'))

    def getRowIndex(self, char):
        '''Gets the index of the row containing this p_char. If p_char is not
           found, return -1.'''
        # Get the PX corresponding to p_char
        px = Layout.pxs[char]
        i = 0
        for row in self.rows:
            for cell in row.cells:
                if px in cell.content:
                    return i
            i += 1
        return -1

    def removeElement(self, elem):
        '''Removes given p_elem from myself'''
        toRemove = Layout.pxs[elem]
        for row in self.rows:
            for cell in row.cells:
                if toRemove in cell.content:
                    cell.content.remove(toRemove)
        if elem in self.layoutString:
            self.layoutString = self.layoutString.replace(elem, '')

    def setRequired(self, required):
        '''Adds or remove the p_required part'''
        if not required:
            self.removeElement('r')
        else:
            # Add it if not already in it
            if 'r' not in self.layoutString:
                # Do not update editLayout.layoutString, too tricky
                content = self.rows[0].cells[0].content
                content.insert(1, Layout.pxs['r'])

    def __repr__(self): return '<Layout %s>' % self.layoutString

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class ColumnLayout:
    '''A "column layout" dictates the way a table column must be rendered. Such
       a layout is of the form: <name>[*width][,|!|`|`][^].'''

    # --------------------------------------------------------------------------
    #   <name>  | is the name of the field whose content must be shown in
    #           | column's cells ;
    # --------------------------------------------------------------------------
    #   width   | is the width of the column. Any valid value for the "width"
    #           | attribute of the "td" HTML tag is accepted ;
    # --------------------------------------------------------------------------
    #   , | !   | column alignment: respectively, left, centered or right ;
    # --------------------------------------------------------------------------
    #     ^     | if present, indicates that the column header must be empty.
    # --------------------------------------------------------------------------

    def __init__(self, layoutString):
        self.layoutString = layoutString

    def get(self):
        '''Returns a list containing the separate elements that are within
           self.layoutString.'''
        consumed = self.layoutString
        lastChar = consumed[-1]
        # Must the column header be empty ?
        if lastChar == '^':
            header = False
            consumed = consumed[:-1]
        else:
            header = True
        # Determine column alignment
        lastChar = consumed[-1]
        if lastChar in Layout.cellDelimiters:
            align = Layout.cellDelimiters[lastChar]
            consumed = consumed[:-1]
        else:
            align = 'left'
        # Determine name and width
        if '*' in consumed:
            name, width = consumed.rsplit('*', 1)
        else:
            name = consumed
            width = ''
        return name, width, align, header

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Layouts(dict):
    '''Define layouts for some field'''

    # When a layouted object must be displayed, depending on the place in the
    # UI where it must appear, a specific layout will be used. This is the
    # concept of "layout type". Appy currently defines the following layout
    # types.
    types = (
      'view',    # a read-only page for an object;
      'edit',    # the editable page for an object;
      'result',  # a cell within a list of objects;
      'query',   # the top-level zone on the page displaying results of some
                 # search, typically used for rendering Pod fields allowing
                 # to export documents including these results;
      'buttons', # the series of buttons rendered on a view/edit page or
                 # besides the object title on a list of objects;
      'xml',     # similar to 'view', but in XML format, readable by a peer
                 # system, the 'view' layout being in XHTML format, readable
                 # by a human being;
      'sidebar'  # The zone in the UI situated, for left-to-right languages,
                 # at the right of the screen.
    )

    # Note that a layout will not be defined for every layout type. For example,
    # layout "xml" renders pure data, so layouting as no sense in that case.
    # Some layouts only define an alternate "place" to render an object, but
    # will use the same rendering as for layout "view". This is the case for
    # layouts "sidebar" and "query" for example.

    # Base types are those for which a Layout instance will be stored in a
    # Layouts instance.
    baseTypes = ('edit', 'view', 'search')

    # Some layouts must be explicitly specified in order to be taken into
    # account, while, for most of them, this is implicit. For example, when
    # determining visibility of a field, boolean "True" implicitly denotes any
    # layout type, excepted those listed in the following attribute.
    explicitTypes = ('buttons', 'sidebar')

    def __init__(self, edit=None, view=None, search=None):
        '''Initialise layouts based on the specified attributes'''
        # Initalise dict keys based on constructor attributes
        if edit: edit = self['edit'] = self.getLayoutFrom(edit)
        if view: view = self['view'] = self.getLayoutFrom(view)
        if search: search = self['search'] = self.getLayoutFrom(search)
        # Derive unspecified layouts from the existing ones, when possible
        if edit and not view:
            view = self['view'] = Layout(other=edit, derivedType='view')
        if view and not search:
            # Do not derive a search layout from a view: because most search
            # forms are similar, use a standard layout else.
            self['search'] = Layout('l-f')

    def getLayoutFrom(self, this):
        '''Gets a Layout instance from p_this'''
        # If p_this is already a Table instance, it is simply returned. If it is
        # a layout string, a Table instance is created from it.
        return this if isinstance(this, Layout) else Layout(this)

    def isComplete(self):
        '''Returns True if p_self contains an entry for each base layout type'''
        for base in Layouts.baseTypes:
            if base not in self:
                return

    def completeFrom(self, other):
        '''Complete p_self by cloning the missing elements from p_other'''
        # p_other is supposed to be complete
        for base in Layouts.baseTypes:
            if base not in self:
                self[base] = Layout(other=other[base])

    def clone(self):
        '''Create a clone from p_self'''
        r = Layouts()
        r.completeFrom(self)
        return r

    def hasPart(self, part):
        '''Return True if the given layout p_part can be found at least once
           among the various p_layouts defined on p_self.'''
        for layout in self.values():
            if part in layout.layoutString: return True

    # ~~~ Default cell layout ~~~
    # Layouts for a field do not contain any entry for layout type "cell".
    # Indeed, rendering a field on such layout is, implicitly, simply rendering
    # its value ("f"). Moreover, rendering a layout renders a complete XHTML
    # table, even if it contains a single cell, like for "f". So we avoid doing
    # this for performance and conciseness reasons. But we can't complelety
    # forget cell layouts, because they are needed for inline-editing (or maybe
    # other Ajax-like functionalities in the future). Indeed, the machinery for
    # ajax-editing the field is inlaid in the XHTML table corresponding to the
    # layout. In such case, we will use a standard cell layout defined here.
    cell = Layout('f|', align='center', css_class='no')

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                              Class methods
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    @classmethod
    def getDefault(class_, field):
        '''Returns the default layouts for this p_field'''
        # ~~~
        # The default layout depends on the type of group into which the field
        # is. This method may be overridden by child Layouts classes as possibly
        # defined on every sub-field class.
        # ~~~
        key = 'grid' if field.inGrid() else 'normal'
        return class_.defaults[key]

    @classmethod
    def getFor(class_, field, layouts):
        '''Gets a Layouts instance for this p_field, based on the value given in
           its "layouts" attribute.'''
        # Get default layouts if p_layouts is empty
        if not layouts:
            r = field.Layouts.getDefault(field)
            defaults = True
        elif isinstance(layouts, str) or isinstance(layouts, Layout):
            # This is a single layout string or instance, the "edit" one
            r = Layouts(layouts)
            defaults = False
        else:
            # p_layouts is already a Layouts instance. Complete it if necessary.
            r = layouts
            if not r.isComplete():
                r.completeFrom(field.Layouts.getDefault(field))
            defaults = False
        # Clone the layout and add special CSS classes in it when appropriate
        if field.focus:
            r = r.clone()
            r['view'].addCssClasses('focus')
            r['edit'].addCssClasses('focus')
        return r

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
#                            Default field layouts
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

# The following layout instances, stored as static attributes on class Layouts,
# are default layouts applying to fields. You can define them as values for
# attributes "layouts" when defining fields on your classes.
# 
# These layouts are generic and apply to most fields. However, every field
# sub-class may define specific layouts. Here are some examples.
# - If you need to use generic layout named "d" as defined hereafter, while
#   defining your field, use
#                             layouts=Layouts.b
# - Class Boolean defines a specific "d" layout that is more appropriate. If you
#   want to use it, write:
#                         layouts=Boolean.Layouts.b
# Note that, because classes <SubField>.Layouts inherits from class Layouts,
# also available as Field.Layouts, if you don't know if there is a specific
# version of some layout on the sub-class, you can systematically use the layout
# as defined on <SubClass>.Layouts. But the notation is longer, so, if you
# prefer a shorter syntax, Layouts.<letter> is available, too.

# If you do not specify any layout at all in the "layouts" attribute for some
# field, default layouts will be applied: either those being defined in the
# following dict, either those as defined in method
#                          <subField>.Layout.getDefaults
# if defined.

# The following naming conventions apply.
# "b"  base layout
# "d"  description (for layouts displaying the field's description)
# "h"  help (for layouts displaying the field's help icon)
# "n"  narrow (for non-100%-wide layouts)
# "w"  wide
# "g"  group (for layouts applicable to fields in groups with style="grid")
# "c"  centered
b = 'lrv-f'
d = 'lrv-d-f'
Layouts.b   = Layouts(Layout(b, width=None))
Layouts.c   = Layouts(Layout('lrv|-f|', width=None, align='center'))
Layouts.d   = Layouts(d)
Layouts.dc  = Layouts(d + '|')
Layouts.h   = Layouts('lhrv-f')
Layouts.n   = Layouts.b
Layouts.nd  = Layouts(Layout(d, width=None))
Layouts.w   = Layouts(Layout(b))
# Layouts with content only (no label)
Layouts.f   = Layouts(Layout('f', width=None))
Layouts.wf  = Layouts(Layout('f'))
Layouts.fv  = Layouts(edit=Layout(b), view=Layout('f'))
Layouts.fvd = Layouts(edit=Layout(d), view=Layout('f'))
Layouts.g   = Layouts('frvl')
# The *d*escription is visible, even on the *v*iew layout
Layouts.dv  = Layouts(edit=d, view='l-d-f')
# The *d*escription is shown on the *v*iew layout but not on *e*dit (=*s*imple)
Layouts.dvs = Layouts(edit=b, view='l-d-f')
# "Grid" group-related layouts
Layouts.gn  = Layouts(Layout('f;rvl=', width=None))
Layouts.gd  = Layouts(Layout('frvl-d', width='99%'))
Layouts.gdn = Layouts(Layout('d2-f;rvl=', width=None))
Layouts.gh  = Layouts('fhrvl')
Layouts.gdh = Layouts('fhrvl-d')
# Base layouts, with space above
Layouts.t = Layouts(Layout(b, width=None, css_class='topSpace'))
Layouts.td = Layouts(Layout(d, css_class='topSpace'))

# Fields being in groups having style "grid" deserve specific layouts. This why
# there are 2 default layouts.
Layouts.defaults = {'normal': Layouts.b, 'grid': Layouts.g}

# Default layouts for pages are defined hereafter. In order to define layouts on
# a Appy class, place an instance of class Layouts (a custom one, or one of the
# layouts as defined hereafter) in static attribute "layouts" on your class.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
#                             Default page layouts
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

class PageLayouts:
    # The default layouts
    defaults = Layouts(edit=Layout('b-w', width=None), view=Layout('e-b-w'))
    # Wide
    wide = Layouts(edit=Layout('b-w'), view=Layout('e-b-w'))
    # Centered
    centered = Layouts(edit=Layout('b|-w|', width=None, align='center'),
                       view=Layout('e-b|-w|', align='center'))
    # Headless (without header on "view")
    headless = Layouts(edit=Layout('b-w', width=None), view=Layout('b-w'))

Layouts.Page = PageLayouts
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.px import Px
from appy.model.base import Base
from appy.ui.layout import Layouts
from appy.database.operators import or_
from appy.model.fields.group import Group
from appy.model.searches.modes import Mode
from appy.model.fields.string import String
from appy.model.fields.integer import Integer
from appy.model.fields.boolean import Boolean
from appy.model.workflow.standard import Owner
from appy.model.searches import Search, UiSearch
from appy.model.fields.select import Select, Selection

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Query(Base):
    '''Persistent and editable parameters for a Search'''

    workflow = Owner
    # Queries are not indexed by default
    indexable = False
    listColumns = ('title', 'className', 'states', 'sortBy', 'sortOrder','mode')
    pageListColumns = ('title', 'state')

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                            Main parameters
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # Fields in the main page are rendered in a grid group
    mainGroup = Group('main', ['250em', ''], style='grid', hasLabel=False)

    qp = {'label': 'Query', 'group': mainGroup}

    # The name of the root class for which instances are searched
    def listRootClasses(self):
        '''Lists the app's root classes'''
        return [(name, self.translate('%s_plural' % name)) \
                for name in self.config.model.rootClasses]

    className = Select(validator=Selection(listRootClasses),
                       multiplicity=(1,1), **qp)

    # Only objects being in one of the states defined in the hereabove field
    # will be part of the result.
    def listStates(self, className=None):
        '''Lists the states from the workflow tied to the class whose name is
           p_className.'''
        className = className or self.className
        r = []
        if not className: return r
        for state in self.model.classes[className].workflow.states.values():
            r.append((state.name, self.translate(state.labelId)))
        return r

    states = Select(validator=Selection(listStates), multiplicity=(1,None),
                    master=className, masterValue=listStates, **qp)

    # Technical indexed fields that the user cannot choose
    unselectableIndexes = ('allowed', 'cid')

    # Field used as sort key
    def listIndexedFields(self, className=None):
        '''Lists, for the class whose name is p_className, the fields being
           indexed.'''
        className = className or self.className
        r = []
        if not className: return r
        for field in self.model.classes[className].fields.values():
            # Ignore the field if it cannot be chose
            if field.name in Query.unselectableIndexes: continue
            if field.indexed:
                r.append((field.name, self.translate(field.labelId)))
        return r

    sortBy = Select(validator=Selection(listIndexedFields), multiplicity=(1,1),
                    master=className, masterValue=listIndexedFields, **qp)
    sortOrder = Select(validator=('asc', 'desc'), multiplicity=(1,1), **qp)
    mode = Select(validator=Mode.concrete, multiplicity=(1,1), **qp)
    maxPerPage = Integer(default=30, multiplicity=(1,1), width=2, **qp)
    showPods = Boolean(default=False, **qp)
    showTitle = Boolean(default=False, **qp)
    showNav = Select(validator=('top', 'bottom', 'both', 'none'),
                     default='both', multiplicity=(1,1), render='radio', **qp)
    navAlign = Select(validator=('left', 'center', 'right'),
                     default='center', multiplicity=(1,1), render='radio', **qp)
    showFilters = Boolean(default=True, **qp)

    def validViaPopup(self, value):
        '''Ensure p_value is valid to be stored in field "viaPopup"'''
        value = value.strip()
        # Value can be "False" or can contain pa popup's width and height, in
        # pixels.
        if not value or (value == 'False'): return True
        value = value.split()
        if len(value) > 2: return self.translate('wrong_via_popup')
        for v in value:
            if not v.endswith('px') or not v[:-2].isdigit():
                return self.translate('wrong_via_popup')
        return True

    viaPopup = String(layouts=Layouts.gd, validator=validViaPopup, **qp)
    pageLayoutOnView = String(layouts=Layouts.gd, **qp)

    def getSearch(self):
        '''Creates the Search and UiSearch instances corresponding to this
           query.'''
        # Get the expression related to p_self.states
        states = self.states
        if len(states) == 1:
            state = states[0]
        else:
            state = or_(states)
        # Get the "viaPopup" parameter
        viaPopup = self.viaPopup or ''
        if viaPopup == 'False':
            viaPopup = False
        elif ' ' in viaPopup: # Popup width and height are specified
            viaPopup = viaPopup.split()
            for i in (0, 1): viaPopup[i] = viaPopup[i].strip()
        elif viaPopup: # Only the popup width is specified
            viaPopup = viaPopup.strip()
        # Create the Search instance corresponding to p_self's attributes
        className = self.className
        class_ = self.model.classes.get(className)
        return Search(name=str(self.iid), maxPerPage=self.maxPerPage,
                      translated=self.translate('%s_plural' % className),
                      sortBy=self.sortBy, sortOrder=self.sortOrder,
                      showPods=self.showPods, showTitle=self.showTitle,
                      showNav=self.showNav, navAlign=self.navAlign,
                      showFilters=self.showFilters, container=class_,
                      viaPopup=viaPopup, pageLayoutOnView=self.pageLayoutOnView,
                      resultModes=(self.mode,), state=state)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                            PX rendering
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # Simulate a HTTP GET on a search corresponding to this Query instance
    pxView = Px('''
     <x var="x=setattr(req, 'className', o.className);
             x=setattr(req, 'search', str(o.iid));
             x=Px.injectRequest(_ctx_, req, tool)">:tool.Search.results</x>''')
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy import Config
from appy.model.base import Base
from appy.model.user import User
from appy.model.fields import Show
from appy.ui.layout import Layouts
from appy.model.fields.ref import Ref
from appy.model.fields.string import String
from appy.model.workflow import standard as workflows
from appy.model.fields.select import Select, Selection
from appy.model.fields.group import Group as FieldGroup

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Group(Base):
    '''Base class representing a group'''
    workflows.Owner

    m = {'group': FieldGroup('main', style='grid', hasLabel=False),
         'width': 25, 'indexed': True, 'layouts': Layouts.g, 'label': 'Group'}

    @staticmethod
    def update(class_):
        title = class_.fields['title']
        title.group = Group.m['group']
        title.layouts = Layouts.g

    def showLogin(self):
        '''When must we show the login field ?'''
        return 'edit' if self.isTemp() else Show.TR

    def showGroups(self):
        '''Only the admin can view or edit roles'''
        return self.user.hasRole('Manager')

    def validateLogin(self, login):
        '''Is this p_login valid ?'''
        return True

    login = String(multiplicity=(1,1), show=showLogin,
                   validator=validateLogin, **m)

    # Field allowing to determine which roles are granted to this group
    roles = Select(validator=Selection(lambda o: o.model.getGrantableRoles(o)),
                   render='checkbox', multiplicity=(0,None), **m)

    users = Ref(User, multiplicity=(0,None), add=False, link='popup',
      height=15, back=Ref(attribute='groups', show=User.showRoles,
                          multiplicity=(0,None), label='User'),
      showHeaders=True, shownInfo=('title', 'login', 'state*100px|'),
      actionsDisplay='inline', label='Group', group=m['group'])
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
try:
    from DateTime import DateTime
except ImportError:
    pass

# Errors -----------------------------------------------------------------------
NON_SSO_USER = 'Non-SSO user "%s" prevents access to the SSO user having the ' \
  'same login.'
INVALID_ROLE = 'Role "%s" mentioned in a HTTP header is not among ' \
  'grantable roles.'
INVALID_GROUP = 'Group "%s" mentioned in a HTTP header was not found.'
SSO_REACTIVATED = '%s reactivated after a new visit via the SSO reverse proxy.'

# ------------------------------------------------------------------------------
class SsoConfig:
    '''When a Appy server is placed behind a reverse proxy performing
       single sign-on (SSO), use this class to tell Appy how and where to
       retrieve, within every HTTP request from the reverse proxy, information
       about the currently logged user: login, groups and roles.'''
    ssoAttributes = {'loginKey': None, 'emailKey':'email', 'nameKey': 'name',
                     'firstNameKey': 'firstName', 'fullNameKey': 'title'}
    otherPrerogative = {'role': 'group', 'group': 'role'}

    # The connection to a SSO server can't really be tested
    testable = False

    # Empty dict
    emptyDict = {}

    def __init__(self):
        # One can give a name to the SSO reverse proxy that will call us
        self.name = ''
        # The HTTP header key that contains the user login
        self.loginKey = ''
        # Keys storing first and last names
        self.firstNameKey = ''
        self.nameKey = ''
        self.fullNameKey = ''
        # Key storing the user email
        self.emailKey = ''
        # If header values are encoded in a specific way, one can provide a
        # function allowing to decode them.
        self.decodeFunction = None
        # Is SSO enabled ?
        self.enabled = True
        # Keys storing user's global roles. If no key is provided, we suppose
        # that roles are not provided by the SSO server and will be managed
        # locally by the Appy app.
        self.roleKey = ''
        # Among (rolesKey, value) HTTP headers, all may not be of interest,
        # for 2 reasons:
        # 1. the "roleKey" may be the same as the "groupKey" (see below);
        # 2. the reverse proxy may send us all roles the currently logged user
        #    has; all those roles may not concern our application.
        # This is why you can define, in "roleRex", a regular expression. The
        # header value will be ignored if this regular expression produces no
        # match. In the case there is a match, note that, if "roleRex" does not
        # have any matching group, the role name will be the complete HTTP
        # header value. Else, the role name will be the first matching group.
        self.roleRex = None
        # For more complex cases, you can define a "roleFunction": a custom
        # fonction that will receive the match object produced by "roleRex" and
        # must return the role name. Funny additional subtlety: if this
        # function returns a tuple (name, None) instead, "name" will be
        # considered a group login, not a role name. This is useful for reverse
        # proxies that send us similar keys for groups and roles.
        self.roleFunction = None
        # Once a role has been identified among HTTP headers via "roleKey" and
        # "roleRex", and has possibly been treated by "roleFunction", its name
        # can be different from the one used in the Appy application. A role
        # mapping can thus be defined.
        self.roleMapping = {}
        # Key storing user's groups. If no key is provided, we suppose
        # that groups are not provided by the SSO server and will be managed
        # locally by the Appy app.
        self.groupKey = ''
        # Regular expression applied to the group value (similar to "roleRex")
        self.groupRex = None
        # If a "group function" is specified, it will receive the match object
        # produced by "groupRex" and must return the group name, or
        # (name, None), similarly to "roleFunction" (in this case, it means
        # that a role name is returned instead of a group login).
        self.groupFunction = None
        # Group mapping (similar to role mapping). Here, we map group logins.
        self.groupMapping = {}
        # If you specify functions in the following attributes, they will be
        # called, respectively, after a user has been linked/unlinked to/from
        # a group. The function receives, as args, the user and the group.
        self.afterLinkGroup = None
        self.afterUnlinkGroup = None
        # For SSO-authenticated users, a specific single-logout URL may be used
        # instead of the classic Appy logout URL.
        self.logoutUrl = None
        # The "synchronization interval" is an integer value representing a
        # number of minutes. Every time a SSO user performs a request on your
        # app, we will not automatically update the corresponding local User
        # instance from HTTP headers' info. We will only do it if this number of
        # minutes has elapsed since the last time we've done it.
        self.syncInterval = 10
        # "encoding" determines the encoding of the header values. Normally, it
        # is defined in header key CONTENT_TYPE. You can force another value
        # here.
        self.encoding = None
        # If the reverse proxy adds some prefix to app URLs, specify it here.
        # For example, if the app is locally available at localhost:8080/ and
        # is publicly available via www.proxyserver.com/myApp, specify "myApp"
        # as URL prefix.
        self.urlPrefix = None
        # In some cases, Appy must know the complete URL allowing to access the
        # app via the reverse proxy. No trailing slash please.
        self.appUrl = None
        # Prior to being deployed behind a SSO server, the Appy app may have
        # managed its users locally. If attribute "convertLocalUsers" is False
        # (the default), when a SSO user having the same login as a local user
        # arrives, Appy will not log him in because a local user exists with
        # this login. If it is set to True, the local user will be automatically
        # converted to a SSO user (excepted if it is a special user: anon,
        # admin,...).
        self.convertLocalUsers = False
        # If you place a function in the following attribute "userOnEdit", it
        # will be called everytime the local copy of a SSO user will be created
        # or updated. Such a local copy is created when a SSO hits the app for
        # the first time; it is then updated on a regular basis, depending on
        # "syncInterval" (see above). "userOnEdit" will be called with 2 args:
        # the User instance and boolean "created", True if the user has just
        # been created or False if it is an update.
        self.userOnEdit = None
        # A direct LDAP connection to the users' directory used by the SSO for
        # authenticating its users can be placed here, as a instance of
        # appy.shared.ldap.LdapConfig.
        self.ldap = None

    def init(self, tool):
        '''Lazy initialisation'''
        # Lazy-initialise the sub-ldap config when present
        if self.ldap:
            self.ldap.init(tool)

    def __repr__(self):
        name = self.name or 'SSO reverse proxy'
        return '%s with login key=%s' % (name, self.loginKey)

    def extractEncoding(self, headers):
        '''What is the encoding of retrieved page?'''
        # Encoding can have a forced value
        if self.encoding: return self.encoding
        if 'CONTENT_TYPE' in headers:
            res = None
            for elem in headers['CONTENT_TYPE'].split(';'):
                elem = elem.strip()
                if elem.startswith('charset='): return elem[8:].strip()
        # This is the default encoding according to HTTP 1.1
        return 'iso-8859-1'

    def getUserParams(self, req):
        '''Formats the user-related data from the request (HTTP headers), as a
           dict of params usable for creating or updating the corresponding
           Appy user.'''
        res = {}
        headers = req._orig_env
        encoding = self.extractEncoding(headers)
        decodeFunction = self.decodeFunction
        for keyAttr, appyName in self.ssoAttributes.iteritems():
            if not appyName: continue
            # Get the name of the HTTP header corresponding to "keyAttr"
            keyName = getattr(self, keyAttr)
            if not keyName: continue
            # Get the value for this header if found in the request
            value = headers.get('HTTP_%s' % keyName, None)
            if value:
                # Apply standard character decoding
                value = value.decode(encoding)
                # Apply specific decoding if any
                if decodeFunction: value = decodeFunction(value)
                res[appyName] = value.encode('utf-8')
        return res

    def extractUserLogin(self, tool, req, warn):
        '''Identify the user from HTTP headers'''
        # Check if SSO is enabled
        if not self.enabled: return
        # Headers could be absent if the request is a fake one at server startup
        headers = getattr(req, '_orig_env', None)
        # If the user we identify exists but as a user from another
        # authentication source:
        # - if it is a local user and self.convertLocalUsers is True, we
        #   convert it to a SSO user;
        # - else, we have a problem: we log a warning and force an
        #   identification failure.
        login = headers and headers.get('HTTP_%s' % self.loginKey, None)
        if login:
            user = tool.search1('User', noSecurity=True, login=login)
            if user and (user.source != 'sso'):
                # Convert it when relevant
                if self.convertLocalUsers:
                    user.convertTo('sso')
                else:
                    # Force an identification failure
                    if warn:
                        tool.log(NON_SSO_USER % login, type='warning',
                                 noUser=True)
                    return
        return login

    def convertUserPrerogatives(self, type, prerogatives, encoding='utf-8'):
        '''Converts SSO p_prerogatives to Appy roles and groups'''
        # p_prerogatives may contain "main" and "secondary" prerogatives.
        # Indeed, when extracting roles we may find groups, and vice versa.
        res = set()
        secondary = set()
        # A comma-separated list of prerogatives may be present
        for value in prerogatives:
            # Ignore empty values
            value = value.strip()
            if not value: continue
            # Get a standardized utf-8-encoded string
            value = value.decode(encoding).encode('utf-8')
            # Apply a regular expression if specified
            rex = getattr(self, '%sRex' % type)
            if rex:
                match = rex.match(value)
                if not match: continue
                # Apply a function if specified
                fun = getattr(self, '%sFunction' % type)
                if fun:
                    value = fun(self, match)
                    if isinstance(value, tuple):
                        # "value" is from the secondary prerogative type
                        value = value[0]
                        # Apply the secondary prerogative mapping if any
                        other = self.otherPrerogative[type]
                        mapping = getattr(self, '%sMapping' % other)
                        if value in mapping:
                            value = mapping[value]
                        secondary.add(value)
                        continue
                else:
                    value = match.group(1)
            # Apply a mapping if specified
            mapping = getattr(self, '%sMapping' % type)
            if value in mapping:
                value = mapping[value]
            # Add the prerogative to the result
            res.add(value)
        return res, secondary
        
    def extractUserPrerogatives(self, type, req):
        '''Extracts, from p_req, user groups or roles, depending on p_type'''
        # Do we care about getting this type or prerogative ?
        key = getattr(self, '%sKey' % type)
        if not key: return None, None
        # Get HTTP headers
        headers = getattr(req, '_orig_env', None)
        if not headers: return None, None
        # Get the HTTP request encoding
        encoding = self.extractEncoding(headers)
        # Extract the value for the "key" header
        headerValue = headers.get('HTTP_%s' % key, None)
        if not headerValue: return None, None
        # We have prerogatives. Convert them to Appy roles and groups.
        return self.convertUserPrerogatives(type, headerValue.split(','),
                                            encoding=encoding)

    def setRoles(self, tool, user, roles):
        '''Grants p_roles to p_user. Ensure those role are grantable.'''
        grantable = tool.o.getProductConfig().grantableRoles
        for role in roles:
            if role in grantable:
                user.addRole(role)
            else:
                tool.log(INVALID_ROLE % role, type='warning', noUser=True)

    def setGroups(self, tool, user, groups):
        '''Puts p_user into p_groups. Ensure those p_groups exist.'''
        for login in groups:
            # Every "login" may represent a single group or, if it contains an
            # asterisc, several groups.
            i = login.find('*')
            if i == -1:
                # "login" is the login of a single group
                group = tool.search1('Group', noSecurity=True, login=login)
                if group:
                    group.link('users', user)
                    if self.afterLinkGroup:
                        self.afterLinkGroup(user, group)
                else:
                    tool.log(INVALID_GROUP % login, type='warning', noUser=True)
            else:
                # "login" is a pattern of groups. Search all groups matching the
                # pattern.
                if i == 0:
                    # Groups share the same prefix. Get them all as basis.
                    attrs = self.emptyDict
                    suffix = login[1:]
                else:
                    # Limit the search to groups having the defined prefix
                    attrs = {'login': login[:i+1]}
                    if i == (len(login) - 1):
                        suffix = ''
                    else:
                        suffix = login[i+1:]
                # Search groups and retain those matching the pattern
                for group in tool.search('Group', noSecurity=True, **attrs):
                    if group.login.endswith(suffix):
                        group.link('users', user)
                        if self.afterLinkGroup:
                            self.afterLinkGroup(user, group)

    def extractUserInfo(self, source, isRequest=True):
        '''By default, this method extracts, from HTTP headers present in the
           request (p_source), user data (name, login...), roles and groups.

           If p_isRequest is False, p_source is an already extracted list of
           prerogatives.'''
        if isRequest:
            req = source
            # Extract simple user attributes
            params = self.getUserParams(req)
            # Extract user global roles and groups
            roles, groups2 = self.extractUserPrerogatives('role', req)
            groups, roles2 = self.extractUserPrerogatives('group', req)
        else:
            params = None
            roles, groups2 = self.convertUserPrerogatives('role', source)
            groups, roles2 = self.convertUserPrerogatives('group', source)
        # Merge found roles
        if roles or roles2:
            if roles and roles2:
                roles = roles.union(roles2)
            elif roles2:
                roles = roles2
        # Merge found groups
        if groups or groups2:
            if groups and groups2:
                groups = groups.union(groups2)
            elif groups2:
                groups = groups2
        return params, roles, groups

    def updateUser(self, tool, req, user):
        '''Updates the local p_user with data from request headers'''
        # The last sync date for this user is now
        user.syncDate = DateTime()
        # Update basic user attributes and recompute the user title
        params, roles, groups = self.extractUserInfo(req)
        for name, value in params.iteritems(): setattr(user, name, value)
        user.updateTitle()
        # Update global roles (when roles are provided by the SSO server)
        if self.roleKey:
            existing = user.roles
            # Remove roles not granted anymore
            for role in existing:
                if not roles or (role not in roles): user.delRole(role)
            # Add roles not already granted
            if roles:
                for role in roles:
                    if role not in existing: user.addRole(role)
        # Update groups (when groups are provided by the SSO server). Unlink any
        # existing group and re-link extracted groups.
        if self.groupKey:
            for group in user.groups:
                group.unlink('users', user)
                if self.afterUnlinkGroup:
                    self.afterUnlinkGroup(user, group)
            self.setGroups(tool, user, groups)
        # Reactivate the user if the local copy was noted as deactivated. We
        # trust SSO.
        if user.state == 'inactive':
            msg = SSO_REACTIVATED % user.login
            user.log(msg, noUser=True)
            user.do('reactivate', noSecurity=True, doHistory=False, comment=msg)

    def createUser(self, tool, login, source, isRequest=True, userParams=None):
        '''Creates the local User copy of a SSO user, from user info as present
           in the request (p_source, if p_isRequest is True), or via an already
           extracted list of prerogatives (p_source) if p_isRequest is False. In
           this latter case, p_userParams is a dict containing user parameters
           (login, name, first name, etc).'''
        # Extract info about the user to create
        params, roles, groups = self.extractUserInfo(source, isRequest)
        if userParams:
            if params: userParams.update(params)
            params = userParams
        user = tool.create('users', noSecurity=True, login=login,
                           source='sso', **params)
        # The last sync date for this user is now
        user.syncDate = DateTime()
        # Set user global roles and groups (if in use)
        if self.roleKey: self.setRoles(tool, user, roles)
        if self.groupKey: self.setGroups(tool, user, groups)
        # Custom User initialization
        if self.userOnEdit: self.userOnEdit(user, created=True)
        if isRequest:
            tool.log('SSO user "%s" (%s) created (1st visit here).' % \
                     (login, user.title), noUser=True)
        return user

    def getUser(self, tool, login, createIfNotFound=True):
        '''Returns a local User instance corresponding to a SSO user'''
        # Check if SSO is enabled
        if not self.enabled: return
        # Do we already have a local User instance for this user ?
        req = tool.request
        user = tool.search1('User', noSecurity=True, login=login)
        if user:
            # Update the user only if we have not done it since "syncInterval"
            # minutes. The user may not have "syncDate" initialized yet.
            mustUpdate = False
            syncDate = user.syncDate
            if not syncDate:
                mustUpdate = True
            else:
                interval = (DateTime() - syncDate) * 1440
                mustUpdate = interval > self.syncInterval
            if mustUpdate:
                self.updateUser(tool, req, user) # Update it from HTTP headers
                # Custom User update
                if self.userOnEdit: self.userOnEdit(user, created=False)
        elif createIfNotFound:
            # Create a local User instance representing the SSO-authenticated
            # user. Collect roles and groups this user has.
            user = self.createUser(tool, login, req)
        return user

    def patchUrl(self, url):
        '''If self.urlPrefix is not empty, check its presence in p_url and
           remove it if found.'''
        if not self.enabled or not self.urlPrefix: return url
        part = '/%s/' % self.urlPrefix
        return url.replace(part, '/', 1)
# ------------------------------------------------------------------------------



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import sys, os, os.path, re, time
from optparse import OptionParser

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
htmlFilters = {'odt': 'HTML (StarWriter)',
               'ods': 'HTML (StarCalc)',
               'odp': 'impress_html_Export'}

FILE_TYPES = {'odt': 'writer8',
              'ods': 'calc8',
              'odp': 'impress8',
              'htm': htmlFilters, 'html': htmlFilters,
              'rtf': 'Rich Text Format',
              'txt': 'Text',
              'csv': 'Text - txt - csv (StarCalc)',
              'pdf': {'odt': 'writer_pdf_Export',  'ods': 'calc_pdf_Export',
                      'odp': 'impress_pdf_Export', 'htm': 'writer_pdf_Export',
                      'html': 'writer_pdf_Export', 'rtf': 'writer_pdf_Export',
                      'txt': 'writer_pdf_Export', 'csv': 'calc_pdf_Export',
                      'swf': 'draw_pdf_Export', 'doc': 'writer_pdf_Export',
                      'xls': 'calc_pdf_Export', 'ppt': 'impress_pdf_Export',
                      'docx': 'writer_pdf_Export', 'xlsx': 'calc_pdf_Export'
                      },
              'swf': 'impress_flash_Export',
              'doc': 'MS Word 97',
              'xls': 'MS Excel 97',
              'ppt': 'MS PowerPoint 97',
              'docx': 'MS Word 2007 XML',
              'xlsx': 'Calc MS Excel 2007 XML',
}
# Conversion from odt to odt does not make any conversion, but updates indexes
# and linked documents.

# Fields of these types are never replaced with their values
unresolvableFieldTypes = ['com.sun.star.text.TextField.PageNumber']
# Complete names for UNO fields that can be "resolved"
resolvableFieldTypes = {'PageCount': 'com.sun.star.text.TextField.PageCount'}
# Table column modifiers
columnModifiers = ('optimalColumnWidths', 'distributeColumns')
# Constants allowing to access LO's stored configuration
configProvider = 'com.sun.star.configuration.ConfigurationProvider'
configAccess = 'com.sun.star.configuration.ConfigurationAccess'
configUpdateAccess = 'com.sun.star.configuration.ConfigurationUpdateAccess'

# Default options when exporting to CSV
defaultCsvOptions = '59,34,76,1'
# Defaut options when reading a CSV file
defaultCsvLoadOptions = defaultCsvOptions

# Constants and messages - - - - - - - - - - - - - - - - - - - - - - - - - - - -
DOC_NOT_FOUND = '"%s" not found.'
URL_NOT_FOUND = 'Doc URL "%s" is wrong. %s'
INPUT_TYPE_ERROR = 'Wrong input type "%s".'
BAD_RESULT_TYPE = 'Bad result type "%s". Available types are %s.'
CANNOT_WRITE_RESULT = 'I cannot write result "%s". %s'
CONNECT_ERROR = "Couldn't not connect to LibreOffice on port %d. %s"
DEFAULT_SERVER = 'localhost'
DEFAULT_PORT = 2002

HELP_SERVER = 'The server IP or hostname that runs LibreOffice ' \
  '(defaults to "%s").' % DEFAULT_SERVER
HELP_PORT = "The port on which LibreOffice runs (default is %d)." % DEFAULT_PORT
HELP_TEMPLATE = 'The path to a LibreOffice template from which you may ' \
  'import styles.'
MANAGE_COLUMNS = 'Set this option to "True" if you want LibreOffice to %s ' \
  'for all tables included in the document. Alternately, specify a regular ' \
  'expression: only tables whose name match will be processed. And if the ' \
  'expression starts with char "~", only tables not matching it will be ' \
  'processed. WARNING - If, for some table, columns are both required to be ' \
  'optimized (parameter "optimalColumnWidths") and distributed (parameter ' \
  '"distributeColumns", only optimization will take place.'
HELP_OPTIMAL_COLUMN_WIDTHS = MANAGE_COLUMNS % 'optimize column widths'
HELP_DISTRIBUTE_COLUMNS = MANAGE_COLUMNS % 'distribute columns evenly'
HELP_SCRIPT = 'You can specify here (the absolute path to) a Python script ' \
  'containing functions that the converter will call in order to customize ' \
  'the process of manipulating the document via the LibreOffice UNO ' \
  'interface. The following functions can be defined in your script and must ' \
  'all accept a single parameter: the Converter instance. ' \
  '***updateTableOfContents***, if defined, will be called for producing a ' \
  'custom table of contents. At the time this function is called by the ' \
  'converter, converter.toc will contain the table of contents, already ' \
  'updated by LibreOffice. ***finalize*** will be called at the end of the '\
  'process, just before saving the result.'
HELP_VERBOSE = 'Writes more information on stdout.'
HELP_PPP = 'Enable the POD post-processor (PPP). The PPP is a series of UNO ' \
  'commands that react to PPP instructions encoded within object names and ' \
  'must be executed at the end of the process, when all other tasks have ' \
  'been performed on the document, just before converting it to another ' \
  'format or saving it to disk or as a stream.'
HELP_STREAM = 'By default (stream = "auto"), if you specify "localhost" as ' \
  'server running LibreOffice, the converter and LibreOffice will exchange ' \
  'the input and result files via the disk. If you specify anything else, ' \
  'the converter will consider that LibreOffice runs on a distant server: ' \
  'input and result files will be carried as streams via the network. If you ' \
  'want to bypass this logic and force exchange as streams or files on disk, ' \
  'set this option to "True" for stream or "False" for disk. You may also ' \
  'specify "in" (the input file is streamed and the result is written on ' \
  'disk) or "out" (the input file is read on disk and the result is streamed).'
HELP_PAGE_START = "Specify an integer number different from 1 and the " \
  "produced document's page numbering will start at this number."
HELP_RESOLVE_FIELDS = 'Set this option to "True" if you want LibreOffice to ' \
  'replace the content of fields by their values. It can be useful, for ' \
  'instance, if the POD result must be included in another document, but the ' \
  'total number of pages must be kept as is. Set this option to "PageCount" ' \
  'instead of "True" to update this field only. Note that field "PageNumber" ' \
  'is never resolved, whatever the value of the option, because its value is ' \
  'different from one page to another.'
HELP_PDF_URL = 'https://wiki.openoffice.org/wiki/API/Tutorials/PDF_export'
HELP_PDF = 'If the output format is PDF, you can define here conversion ' \
  'options, as a series of comma-separated key=value pairs, as in ' \
  '"ExportNotes=True,PageRange=1-20". Available options are documented in ' \
  '%s.' % HELP_PDF_URL
HELP_CSV_URL = 'https://wiki.openoffice.org/wiki/Documentation/DevGuide/' \
  'Spreadsheets/Filter_Options#Filter_Options_for_the_CSV_Filter'
HELP_CSV = 'If the ouput format is CSV, you can define here conversion ' \
  'options, as a comma-separated list of values. Default options are: %s.' \
  'Values correspond to ASCII codes. The first one represents the field ' \
  'separator. The most frequent values are: 59 (the semi-colon ;), 44 (the ' \
  'comma ,) and 9 (a tab). The second value represents the text delimiter. ' \
  'The most frequent values are: 34 (double quotes), 39 (single quotes) or ' \
  'no value at all (as in 59,,76,1). The third one is the file encoding. The ' \
  'most frequent values are 76 (UTF-8) and 12 (ISO-8859-1). Complete ' \
  'documentation about CSV options can be found at %s.' % \
  (defaultCsvOptions, HELP_CSV_URL)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Classes defined here are used to contact LibreOffice via input/output streams
# instead of via the file system. In use when Libreoffice runs on a distant
# server (converter.server is not "localhost").
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
try: # UNO may not be there
    import uno
    from unohelper import Base
    from com.sun.star.io import XInputStream, XOutputStream, XSeekable

    class InputStream(Base, XInputStream, XSeekable):
        def __init__(self, path):
            self.stream = open(path, 'rb')
            self.stream.seek(0, os.SEEK_END)
            self.size = self.stream.tell()
        def readBytes(self, retSeq, nByteCount):
            retSeq = self.stream.read(nByteCount)
            return (len(retSeq), uno.ByteSequence(retSeq))
        def readSomeBytes(self, foo, n):
            return self.readBytes(foo, n)
        def skipBytes(self, n):
            self.stream.seek(n, 1)
        def available(self):
            return self.size - self.stream.tell()
        def closeInput(self):
            self.stream.close()
        def seek(self, posn):
            self.stream.seek(int(posn))
        def getPosition(self):
            return self.stream.tell()
        def getLength(self):
            return self.size

    class OutputStream(Base, XOutputStream):
        def __init__(self, path):
            self.closed = 0
            self.stream = open(path, 'wb')
        def closeOutput(self):
            self.closed = 1
            self.stream.close()
        def writeBytes(self, seq):
            self.stream.write(seq.value)
        def flush(self):
            pass
except ImportError:
    pass

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class LoIter:
    '''Iterates over a collection of LibreOffice-UNO objects'''
    def __init__(self, collection, log=None, reverse=False):
        self.collection = collection
        self.count = collection.getCount()
        # Log the number of walked elements when relevant
        if log: log(self.count, cr=False)
        self.reverse = reverse
        if reverse:
            self.i = self.count - 1
        else:
            self.i = 0
        from com.sun.star.lang import IndexOutOfBoundsException
        self.error = IndexOutOfBoundsException

    def __iter__(self): return self

    def __next__(self):
        try:
            elem = self.collection.getByIndex(self.i)
            if self.reverse: self.i -= 1
            else: self.i += 1
            return elem
        except self.error:
            # IndexOutOfBoundsException can be raised because sometimes UNO
            # returns a number higher than the real number of elements. This is
            # because, in documents including sub-documents, it also counts the
            # sections that are present within these sub-documents.
            raise StopIteration

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class PPP:
    '''This class represents the POD post-processor (=PPP), a series of UNO
       commands that execute instructions inlaid within element names.'''

    def __init__(self, doc):
        # The document currently opened within LO
        self.doc = doc

    def updateTables(self):
        '''When a table name has a suffix between brackets, it means that some
           rows must be removed. This method performs these removals.'''
        for table in LoIter(self.doc.getTextTables()):
            name = table.Name
            # Ignore irrelevant tables
            if '(' not in name: continue
            rows = table.getRows()
            count = rows.getCount()
            # Extract the "slice"
            slice = name[name.index('(')+1 : name.index(')')]
            if slice.startswith(':'):
                # Keep the starting rows, until "i" (delete ending rows)
                i = int(slice[1:])
                rows.removeByIndex(i, count - i)
            else:
                # Keep the ending rows, from "i" (delete starting rows)
                i = int(slice[:-1])
                rows.removeByIndex(0, i)

    def run(self):
        '''Find and execute all post-processing commands'''
        self.updateTables()

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class PdfOptions:
    # Some PDF options' values must be converted
    values = {'true': True, 'True': True, 'false': False, 'False': False}

    @classmethod
    def get(class_, options):
        '''Get and convert PDF options to a dict'''
        if not options: return
        r = {}
        for option in options.split(','):
            if not option: continue
            elems = option.split('=')
            if len(elems) != 2: continue
            key, value = elems
            if not key or not value: continue
            # Convert value when relevant
            if value in class_.values:
                value = class_.values[value]
            elif value.isdigit():
                value = int(value)
            # Add the final value to the result
            r[key] = value
        return r

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Converter:
    '''Converts a document readable by LibreOffice into pdf, doc, txt, rtf...'''

    # Converter-specific error class
    class Error(Exception): pass

    def __init__(self, docPath, result, server=DEFAULT_SERVER,
                 port=DEFAULT_PORT, templatePath=None, optimalColumnWidths=None,
                 distributeColumns=None, script=None, resolveFields=False,
                 pdfOptions=None, csvOptions=defaultCsvOptions, ppp=False,
                 stream='auto', pageStart=1, verbose=False):
        # The server and port where LibreOffice listens
        self.server = server
        self.port = port
        # The path to the document to convert
        self.docUrl, self.docPath = self.getFilePath(docPath)
        self.inputType = self.getInputType(docPath)
        # "resultType" determines the type of the converted file (=a file
        # extension among FILE_TYPES keys). "resultMultiple" is True if several
        # result files will be produced (like when producing one CSV file for
        # every sheet from an Excel file).
        self.resultType, self.resultMultiple = self.getResultType(result)
        self.resultFilter = self.getResultFilter()
        self.resultUrl = self.getResultUrl(result)
        # LibreOffice (LO) run-time objects
        self.context = None # The LO context
        self.oo = None # The LO application object
        self.doc = None # The document that will be loaded in LO
        self.toc = None # The table of contents, if existing
        # LO version, as a string
        self.version = None
        # The path to a LO template (ie, a ".ott" file) from which styles can be
        # imported.
        self.templateUrl = self.templatePath = None
        if templatePath:
            self.templateUrl, self.templatePath = self.getFilePath(templatePath)
        # Actions to perform on table columns
        # ~~~ Optimal column widths ~~~
        self.optimalColumnWidths = optimalColumnWidths
        # ~~~ Distribute columns evenly ~~~
        self.distributeColumns = distributeColumns
        # ~~~ Must we perform one of the above actions on table columns ? ~~~
        self.processColumns = optimalColumnWidths or distributeColumns
        # Must we replace fields with their values ?
        self.resolveFields = resolveFields
        # Functions extracted from the custom Python script that will be called
        # by the Converter for customizing the result.
        if script:
            self.functions = self.getCustomFunctions(script)
        else:
            self.functions = {}
        # Options for conversion to PDF
        self.pdfOptions = PdfOptions().get(pdfOptions)
        # Options for conversion to CSV
        self.csvOptions = csvOptions
        # Verbosity
        self.verbose = verbose
        # Enable the PPP (POD post-processor) ?
        self.ppp = ppp
        # Exchange template and result with LibreOffice via streams or files
        # on disk ?
        self.stream = stream
        # At what page number must the produced document start ?
        self.pageStart = pageStart

    def log(self, msg, cr=True):
        '''Logs some p_msg if we are in verbose mode'''
        if not self.verbose: return
        if not isinstance(msg, str): msg = str(msg)
        sys.stdout.write(msg)
        if cr:
            sys.stdout.write('\n')

    def getInputType(self, docPath):
        '''Extracts the input type from the p_docPath'''
        res = os.path.splitext(docPath)[1][1:].lower()
        if res not in FILE_TYPES: raise self.Error(INPUT_TYPE_ERROR % res)
        return res

    def getResultType(self, result):
        '''p_result can be the aobsolute path to the result to produce or only a
           file extension indicating the format of the file to produce. This
           method returns this format.'''
        if '.' in result:
            # We have a complete path
            r = os.path.splitext(result)[-1][1:]
        else:
            r = result
        # If result type ends with char '*', it means that several output files
        # will be produced, ie for getting a CSV file from every Excel or Calc
        # sheet.
        if r.endswith('*'): return r[:-1], True
        return r, False

    def getCustomFunctions(self, script):
        '''Compiles and executes the Python p_script and returns a dict
           containing its namespace.'''
        f = open(script)
        content = f.read()
        f.close()
        names = {}
        exec(compile(content, script, 'exec'), names)
        return names

    def getFilePath(self, filePath):
        '''Returns the absolute path of p_filePath. In fact, it returns a
           tuple with some URL version of the path for LO as the first element
           and the absolute path as the second element.''' 
        import unohelper
        if not os.path.exists(filePath) and not os.path.isfile(filePath):
            raise self.Error(DOC_NOT_FOUND % filePath)
        docAbsPath = os.path.abspath(filePath)
        # Return one path for OO, one path for me
        return unohelper.systemPathToFileUrl(docAbsPath), docAbsPath

    def getResultFilter(self):
        '''Based on the result type, identifies which OO filter to use for the
           document conversion.'''
        if self.resultType in FILE_TYPES:
            res = FILE_TYPES[self.resultType]
            if isinstance(res, dict):
                res = res[self.inputType]
        else:
            raise self.Error(BAD_RESULT_TYPE % (self.resultType,
                                                FILE_TYPES.keys()))
        return res

    def getResultUrl(self, result):
        '''Returns the path of the result file in the format needed by LO. If
           p_result is a file path, it will be used as base. Else, it is a file
           extension (=the "result type"): computing the complete path will be
           done as follows.

           If the result type and the input type are the same (ie the user wants
           to refresh indexes or some other action and not perform a real
           conversion), the result file will be named
                           <inputFileName>.res.<resultType>

           Else, the result file will be named like the input file but with a
           different extension:
                           <inputFileName>.<resultType>
        '''
        import unohelper
        if '.' in result:
            # It is a full path
            r = result
        else:
            # Name the result based on the input file
            baseName = os.path.splitext(self.docPath)[0]
            if self.resultType != self.inputType:
                r = '%s.%s' % (baseName, self.resultType)
            else:
                r = '%s.res.%s' % (baseName, self.resultType)
        try:
            f = open(r, 'w')
            f.write('Hello')
            f.close()
            os.remove(r)
            return unohelper.systemPathToFileUrl(r)
        except (OSError, IOError) as e:
            raise self.Error(CANNOT_WRITE_RESULT % (r, e))

    def props(self, properties):
        '''Create a UNO-compliant tuple of properties, from tuple p_properties
           containing sub-tuples (s_propertyName, value).'''
        from com.sun.star.beans import PropertyValue
        res = []
        for name, value in properties:
            prop = PropertyValue()
            prop.Name = name
            prop.Value = value
            res.append(prop)
        return tuple(res)

    def getConfig(self, path, update=False):
        '''Returns a node within the LO configuration, stored at p_path. If the
           objective is to update it, set p_update to True.'''
        provider = self.context.ServiceManager.createInstance(configProvider)
        prop = self.props((('nodepath', path),))
        access = update and configUpdateAccess or configAccess
        return provider.createInstanceWithArguments(access, prop)

    def getVersion(self):
        '''Returns the LO version'''
        try:
            path = '/org.openoffice.Setup/Product'
            return self.getConfig(path).getByName('ooSetupVersion')
        except Exception:
            # LibreOffice 3 raises an exception here
            return '3.0'

    def connect(self):
        '''Connects to LibreOffice'''
        if os.name == 'nt':
            import socket
        import uno
        from com.sun.star.connection import NoConnectException
        try:
            # Get the uno component context from the PyUNO runtime
            self.context = ctx = uno.getComponentContext()
            # Get the LO version
            self.version = self.getVersion()
            # Create the UnoUrlResolver
            create = ctx.ServiceManager.createInstanceWithContext
            resolver = create('com.sun.star.bridge.UnoUrlResolver', ctx)
            # Connect to LO running on self.port
            docContext = resolver.resolve(
              'uno:socket,host=%s,port=%d;urp;StarOffice.ComponentContext' % \
              (self.server, self.port))
            # Is seems that we can't define a timeout for this method. This
            # would be useful because when a non-LO server already listens
            # to self.port, this method blocks.
            self.log('Getting the UNO-LO instance...', cr=False)
            self.oo = docContext.ServiceManager.createInstanceWithContext(
                'com.sun.star.frame.Desktop', docContext)
            # If we must process table column widths, create a dispatch helper
            if self.processColumns:
                helper = create('com.sun.star.frame.DispatchHelper', ctx)
                self.dispatchHelper = helper
            self.log(' done.')
        except NoConnectException as e:
            raise self.Error(CONNECT_ERROR % (self.port, e))

    def getColumnModifiers(self):
        '''Returns the elements allowing to know if we must optimize or
           distribute table columns.'''
        r = []
        add = r.append
        for name in columnModifiers:
            value = getattr(self, name)
            if isinstance(value, bool):
                # Action must be performed on all tables or no table at all
                add(value); add(None)
            elif value:
                # Action must be performed on tables whose name match or not
                # some regular expression
                if value.startswith('~'):
                    value = value[1:]
                    vNot = True
                else:
                    vNot = False
                add(re.compile(value)); add(vNot)
            else:
                add(False); add(None)
        return r

    def processTableColumns(self, table, viewCursor, frame,
                            optimize, distribute):
        '''Process column widths for this p_table: either p_optimize them or
           p_distribute them evenly (one of these 2 params must be True). Note
           that:
           * the "optimize" algorithm dos not work properly in all cases, unless
             algorithm "distribute" is executed beforehand. So, in short:
                      optimize = distribute + optimize
           * if both p_optimize and p_distribute are specified, the p_optimize
             operation will be performed and will produce the final result.
             Indeed, even if a "distribute" preamble operation is performed, it
             will not be the final result.
        '''
        cursor = table.getCellByName('A1').createTextCursor()
        viewCursor.gotoRange(cursor, False)
        viewCursor.gotoEnd(True)
        viewCursor.gotoEnd(True)
        do = self.dispatchHelper.executeDispatch
        # Perform the "distribute" operation in all cases
        do(frame, '.uno:DistributeColumns', '', 0, ())
        # Perform the "optimize" operation when relevant
        if optimize:
            # With LibreOffice < 5, range selection must be done again
            if self.version < '5.0':
                viewCursor.gotoRange(cursor, False)
                viewCursor.gotoEnd(True)
                viewCursor.gotoEnd(True)
            do(frame, '.uno:SetOptimalColumnWidth', '', 0, ())

    def updateToc(self):
        '''Update the table of contents within the ODT document (if any)'''
        if self.toc is None: return
        self.toc.update()
        # Call custom code to update the TOC when relevant
        fun = self.functions.get('updateTableOfContents')
        if fun: fun(self)

    def mustStream(self, direction):
        '''Must the template and result files be exchanged between the converter
           and LibreOffice as streams over the network or files on disk ?'''
        stream = self.stream
        if stream == 'auto':
            # By default, use streams when LibreOffice runs on a distant
            # server = when the server is not "localhost".
            return self.server != 'localhost'
        elif stream in ('in', 'out'):
            # Stream only in the indicated direction
            return direction == stream
        else:
            # Streaming or not has been forced
            return stream

    def mustBeResolved(self, field):
        '''Must this p_field be replaced with its content ?'''
        # Check that this field is not among unresolvable field types
        for name in unresolvableFieldTypes:
            if field.supportsService(name):
                return
        resolve = self.resolveFields
        if isinstance(resolve, str):
            # We must resolve only fields of a given type
            name = resolvableFieldTypes.get(resolve)
            if not name: return
            return field.supportsService(name)
        else:
            return resolve

    def setPageStart(self):
        '''Sets, for the produced document, a page start different from 1, as
           specified in self.pageStart.'''
        text = self.doc.getText()
        textCursor = text.createTextCursor()
        textCursor.gotoEndOfParagraph(False)
        textCursor.BreakType = 'PAGE_BEFORE'
        textCursor.PageDescName = 'Standard'
        textCursor.PageNumberOffset = self.pageStart

    def updateOdtDocument(self):
        '''If the input file is an ODT document, we will perform those tasks:
           1) update all indexes;
           2) update sections (if sections refer to external content, we try to
              include the content within the result file);
           3) update table column's widths when relevant;
           4) load styles from an external template if given;
           5) execute PPP commands when relevant.
        '''
        log = self.log
        # The produced document may start at a page number which is not 1
        if self.pageStart != 1: self.setPageStart()
        # Getting some base LO objects is required
        controller = self.doc.getCurrentController()
        viewCursor = controller.getViewCursor()
        # The following lines of code were a trial to force LO to load big
        # documents in their entirery: because, with some early LO 6.x versions,
        # it could happen that resulting PDF results were truncated. But in some
        # rare cases, those lines caused LO to run 100% CPU in an infinite loop.
        # ~~~
        # viewCursor.jumpToLastPage()
        # viewCursor.jumpToEndOfPage()
        # ~~~
        # 1) Update indexes
        for index in LoIter(self.doc.getDocumentIndexes(), log):
            if index.ServiceName == 'com.sun.star.text.ContentIndex':
                # Update the TOC later, after any other action that may still
                # produce changes in page numbers.
                self.toc = index
            else:
                index.update()
        log(' index(es) updated.')
        # 2) Resolve (some) text fields when relevant
        if self.resolveFields:
            for field in self.doc.TextFields:
                # Ignore fields that must not be resolved
                if not self.mustBeResolved(field): continue
                anchor = field.getAnchor()
                cursor = anchor.getText().createTextCursorByRange(anchor)
                cursor.setString(cursor.getString())
            log(' field(s) updated.')
        # 3) Update sections
        self.doc.updateLinks()
        for section in LoIter(self.doc.getTextSections(), log, reverse=True):
            # I must walk into the section from last one to the first one. Else,
            # when "disposing" sections, I remove sections and the remaining
            # ones get another index.
            if section.FileLink and section.FileLink.FileURL:
                # This call removes the <section></section> tags without
                # removing the content of the section. Else, it won't appear.
                section.dispose()
        log(' section(s) updated.')
        # 4) Update tables
        count = 0
        if self.processColumns:
            optimize, oNot, distribute, dNot = self.getColumnModifiers()
            # Browse tables
            frame = controller.getFrame()
            for table in LoIter(self.doc.getTextTables()):
                # Must columns be optimized or distributed for this table ?
                opt = optimize
                if not isinstance(opt, bool):
                    opt = opt.match(table.Name)
                    if oNot: opt = not opt
                dis = distribute
                if not isinstance(dis, bool):
                    dis = dis.match(table.Name)
                    if dNot: dis = not dis
                if opt or dis:
                    self.processTableColumns(table, viewCursor, frame, opt, dis)
                    count += 1
        log('%d table(s) with optimized widths.' % count)
        # 5) Import styles from an external file when required
        if self.templateUrl:
            params = self.props((('OverwriteStyles', True),
                                 ('LoadPageStyles', False)))
            self.doc.StyleFamilies.loadStylesFromURL(self.templateUrl, params)
            log('Styles loaded from %s.' % self.templateUrl)
        # 6) Execute PPP commands when relevant
        if self.ppp:
            PPP(self.doc).run()

    def loadDocument(self, fileUrl, props):
        '''Loads the document in LO, either as a file or as a stream and returns
           a pointer to it.'''
        if self.mustStream('in'):
            # Create an input stream
            stream = InputStream(fileUrl[7:]) # Remove the leading "file://"
            # Add, among p_props, a new property with the input stream
            props = self.props((('InputStream', stream),)) + props
            # Specify that the stream must be read instead of p_fileUrl
            fileUrl = 'private:stream'
        return self.oo.loadComponentFromURL(fileUrl, "_blank", 0, props)

    def storeDocument(self, resultUrl, doc, props):
        '''Saves the result, either by dumping the result on disk or by
           receiving it from LO as a stream from the network.'''
        if self.mustStream('out'):
            # Create an output stream
            output = OutputStream(resultUrl[7:]) # Remove the leading "file://"
            # Add, among p_props, a new property with the output stream
            props = self.props((('OutputStream', output),)) + props
            # Specify that the stream must be written instead of p_resultUrl
            resultUrl = 'private:stream'
        doc.storeToURL(resultUrl, props)

    def openDocument(self):
        from com.sun.star.lang import IllegalArgumentException
        try:
            # Loads the document to convert in a new hidden frame
            self.log('Loading in LO file %s...' % self.docUrl, cr=False)
            props = [('Hidden', True)]
            if self.inputType == 'csv':
                # Give some additional params if we need to open a CSV file
                props.append(('FilterFlags', defaultCsvLoadOptions))
            self.doc = self.loadDocument(self.docUrl, self.props(props))
            self.log(' done.')
            # [ODT] Perform additional tasks
            isOdt = self.inputType == 'odt'
            if isOdt: self.updateOdtDocument()
            try:
                self.doc.refresh()
            except AttributeError:
                pass
            # [ODT] After every change has been applied, refresh the TOC
            if isOdt: self.updateToc()
        except IllegalArgumentException as e:
            raise self.Error(URL_NOT_FOUND % (self.docPath, e))

    def setPdfOptions(self):
        '''If p_self.doc must be converted to PDF, and PDF options need to be
           applied, a data structure containing these options is r_eturned.'''
        if not self.pdfOptions: return
        import uno
        # Available PDF options are documented @HELP_PDF_URL (see above)
        r = self.props(self.pdfOptions.items())
        return uno.Any("[]com.sun.star.beans.PropertyValue", r)
        # Getting the LO's PDF configuration to update it does not work (tested
        # with LO 5.3).
        #path = '/org.openoffice.Office.Common/Filter/PDF/Export/'
        #config = self.getConfig(path, update=True)
        #config.setPropertyValue('ExportNotes', True)
        # Or: config.ExportNotes = True
        #config.commitChanges()

    def convertDocument(self):
        '''Calls LO to perform a document conversion. Note that the conversion
           is not really done if the source and target documents have the same
           type.'''
        self.log('Saving the result in %s...' % self.resultUrl, cr=False)
        props = [('FilterName', self.resultFilter), ('UpdateDocMode', 3)]
        if self.resultType == 'csv':
            # Add options for CSV export
            props.append(('FilterOptions', self.csvOptions))
        elif self.resultType == 'pdf':
            # Get optional PDF options
            options = self.setPdfOptions()
            if options:
                props.append(('FilterData', options))
        if not self.resultMultiple:
            self.storeDocument(self.resultUrl, self.doc, self.props(props))
            self.log(' done.')
        else:
            # Dump one CSV file for every sheet in the input document
            import unicodedata
            doc = self.doc
            sheets = self.doc.getSheets()
            sheetsCount = sheets.getCount()
            controller = doc.getCurrentController()
            props = self.props(props)
            for i in range(sheetsCount):
                sheet = sheets.getByIndex(i)
                # Compute the csv output file name
                name = unicodedata.normalize('NFKD', sheet.getName())
                splitted = os.path.splitext(self.resultUrl)
                resultUrl = '%s.%s%s' % (splitted[0], name, splitted[1])
                controller.setActiveSheet(sheet)
                self.storeDocument(resultUrl, doc, props)

    def run(self):
        '''Connects to LO, does the job and disconnects'''
        if self.verbose: start = time.time()
        self.connect()
        self.openDocument()
        # Call custom code to modify the document when relevant
        fun = self.functions.get('finalize')
        if fun: fun(self)
        # Store the (converted) result
        self.convertDocument()
        self.doc.close(True)
        if self.verbose:
            self.log('Done in %.2f second(s).' % (time.time() - start))

# ConverterScript constants  - - - - - - - - - - - - - - - - - - - - - - - - - -
WRONG_NB_OF_ARGS = 'Wrong number of arguments.'
ERROR_CODE = 1
usage = '''usage: python3 converter.py fileToConvert output [options]

   "fileToConvert" is the absolute or relative pathname of the file you
   want to convert (or whose content like indexes need to be refreshed)
   
   "output" can be the output format, that must be one of: %s
            or can be the absolute path to the result file, whose extension must
            correspond to a valid output format.

   "python" should be a UNO-enabled Python interpreter (ie the one which is
   included in the LibreOffice distribution).''' % str(FILE_TYPES.keys())

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class ConverterScript:
    '''The command-line program'''

    def run(self):
        optParser = OptionParser(usage=usage)
        add = optParser.add_option
        add('-e', '--server', dest='server', default=DEFAULT_SERVER,
            metavar='SERVER', type='string', help=HELP_SERVER)
        add('-p', '--port', dest='port', default=DEFAULT_PORT,
            metavar='PORT', type='int', help=HELP_PORT)
        add('-t', '--template', dest='template', default=None,
            metavar='TEMPLATE', type='string', help=HELP_TEMPLATE)
        add('-o', '--optimalColumnWidths', dest='optimalColumnWidths',
            default=None, metavar='OPTIMAL_COL_WIDTHS', type='string',
            help=HELP_OPTIMAL_COLUMN_WIDTHS)
        add('-d', '--distributeColumns', dest='distributeColumns',
            default=None, metavar='DISTRIBUTE_COLUMNS', type='string',
            help=HELP_DISTRIBUTE_COLUMNS)
        add('-r', '--resolveFields', dest='resolveFields', default=None,
            metavar='RESOLVE_FIELDS', type='string', help=HELP_RESOLVE_FIELDS)
        add('-s', '--script', dest='script', default=None, metavar='SCRIPT',
            type='string', help=HELP_SCRIPT)
        add('-v', '--verbose', action='store_true', help=HELP_VERBOSE)
        add('-f', '--pdf', dest='pdf', default=None, metavar='PDF_OPTIONS',
            type='string', help=HELP_PDF)
        add('-i', '--csv', dest='csv', default=defaultCsvOptions,
            metavar='CSV_OPTIONS', type='string', help=HELP_CSV)
        add('-c', '--ppp', action='store_true', help=HELP_PPP)
        add('-a', '--stream', dest='stream', default='auto',
            metavar='STREAM', type='string', help=HELP_STREAM)
        add('-g', '--pageStart', dest='pageStart', default=1,
            metavar='PAGESTART', type='int', help=HELP_PAGE_START)
        options, args = optParser.parse_args()
        if len(args) != 2:
            sys.stderr.write(WRONG_NB_OF_ARGS)
            sys.stderr.write('\n')
            optParser.print_help()
            sys.exit(ERROR_CODE)
        # Apply relevant type conversions to options
        optimize = options.optimalColumnWidths
        if optimize in ('True', 'False'): optimize = eval(optimize)
        distribute = options.distributeColumns
        if distribute in ('True', 'False'): distribute = eval(distribute)
        resolveFields = options.resolveFields
        if resolveFields == 'True': resolveFields = True
        stream = options.stream
        if stream in ('True', 'False'): stream = eval(stream)
        converter = Converter(args[0], args[1], options.server, options.port,
          options.template, optimize, distribute, options.script, resolveFields,
          options.pdf, options.csv, options.ppp, stream, options.pageStart,
          options.verbose)
        try:
            converter.run()
        except Converter.Error as err:
            sys.stderr.write(str(err))
            sys.stderr.write('\n')
            optParser.print_help()
            sys.exit(ERROR_CODE)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
if __name__ == '__main__':
    ConverterScript().run()
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import re

from appy.xml import Tag
from appy.pod.elements import *
from appy.pod.buffers import FileBuffer, MemoryBuffer
from appy.pod.odf_parser import OdfEnvironment, OdfParser

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class OdTable:
    '''Informations about the currently parsed Open Document (Od)table.'''
    def __init__(self):
        self.nbOfColumns = 0
        self.nbOfRows = 0
        self.curColIndex = None
        self.curRowAttrs = None

    def isOneCell(self):
        return (self.nbOfColumns == 1) and (self.nbOfRows == 1)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class OdInsert:
    '''While parsing an odt/pod file, we may need to insert a specific odt chunk
       at a given place in the odt file (ie: add the pod-specific fonts and
       styles). OdInsert instances define such 'inserts' (what to insert and
       when).'''
    def __init__(self, odtChunk, tag, nsUris={}):
        # The odt chunk to insert
        self.odtChunk = odtChunk
        # The p_odtChunk will be inserted just after the p_tag starts. If more
        # than one p_tag is present in the odt file, the p_odtChunk will be
        # inserted only at the first p_tag occurrence.
        self.tag = tag

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class PodEnvironment(OdfEnvironment):
    '''Contains all elements representing the current parser state during
       parsing.'''
    # Possibles modes
    # ADD_IN_BUFFER: when encountering an impactable tag, we must continue to
    #                dump it in the current buffer.
    ADD_IN_BUFFER = 0
    # ADD_IN_SUBBUFFER: when encountering an impactable tag, we must create a
    #                   new sub-buffer and dump it in it.
    ADD_IN_SUBBUFFER = 1
    # Possible states
    IGNORING = 0 # We are ignoring what we are currently reading
    READING_CONTENT = 1 # We are reading "normal" content
    READING_STATEMENT = 2 # We are reading a POD statement (for, if...)
    READING_EXPRESSION = 3 # We are reading a POD expression.
    # Tags that will be read within notes
    NOTE_TAGS = ('text:p', 'text:span')
    # Start and end tags for every expression holder
    exprStartTags = {
      'if': 'text:conditional-text', 'change': 'text:change-start',
      'input': 'text:text-input', 'db': 'text:database-display'}
    exprEndTags = {
      'if': 'text:conditional-text', 'change': 'text:change-end',
      'input': 'text:text-input', 'db': 'text:database-display'}

    def __init__(self, context, inserts, expressionsHolders):
        OdfEnvironment.__init__(self)
        # Buffer where we must dump the content we are currently reading
        self.currentBuffer = None
        # XML element content we are currently reading
        self.currentContent = ''
        # Current statement (a list of lines) that we are currently reading
        self.currentStatement = []
        # Current mode
        self.mode = self.ADD_IN_SUBBUFFER
        # Current state
        self.state = self.READING_CONTENT
        # Elements we must ignore (they will not be included in the result)
        self.ignorableTags = None # Will be set after namespace propagation
        # Elements that may be impacted by POD statements
        self.impactableTags = None # Idem
        # Elements representing start and end tags surrounding expressions
        self.exprStartElems = self.exprEndElems = None # Idem
        # Stack of currently visited tables
        self.tableStack = []
        self.tableIndex = -1
        # Evaluation context
        self.context = context
        # For the currently read expression, is there style-related information
        # associated with it?
        self.exprHasStyle = False
        # Namespace definitions are not already encountered.
        self.gotNamespaces = False
        # Store inserts
        self.inserts = inserts
        # Currently walked "if" actions
        self.ifActions = []
        # Currently walked named "if" actions
        self.namedIfActions = {} #~{s_statementName: appy.pod.actions.If}~
        # Currently parsed expression within an ODS template
        self.currentOdsExpression = None
        self.currentOdsHook = None
        # Currently parsed expression from a database-display field
        self.currentDbExpression = None
        # Names of some tags, that we will compute after namespace propagation
        self.tags = None
        # When an error occurs, must we raise it or write it into he current
        # buffer?
        self.raiseOnError = None # Will be initialized by PodParser.__init__
        # The expressions holders in use
        self.expressionsHolders = expressionsHolders

    def getTable(self):
        '''Gets the currently parsed table'''
        res = None
        if self.tableIndex != -1:
            res = self.tableStack[self.tableIndex]
        return res

    def transformInserts(self):
        '''Now the namespaces were parsed; p_inserts can be set in the form of a
           dict for easier and more performant access while parsing.'''
        r = {}
        if not self.inserts: return r
        for insert in self.inserts:
            tagName = insert.tag.getFullName(self.namespaces)
            if tagName not in r:
                r[tagName] = insert
        return r

    def manageInserts(self, tag):
        '''We just dumped the start of an elem. Here we will insert any odt
           chunk if needed.'''
        if tag in self.inserts:
            insert = self.inserts[tag]
            self.currentBuffer.write(insert.odtChunk)
            # The insert is destroyed after single use
            del self.inserts[tag]

    def onStartElement(self, tag, attrs):
        ns = self.namespaces
        if not self.gotNamespaces:
            # We suppose that all the interesting (from the POD point of view)
            # XML namespace definitions are defined at the root XML element.
            # Here we propagate them in XML element definitions that we use
            # throughout POD.
            self.gotNamespaces = True
            self.propagateNamespaces()
        tableNs = self.ns(self.NS_TABLE)
        if tag == Table.OD.nsName:
            self.tableStack.append(OdTable())
            self.tableIndex += 1
        elif tag == Row.OD.nsName:
            table = self.getTable()
            table.nbOfRows += 1
            table.curColIndex = -1
            table.curRowAttrs = attrs
        elif tag == Cell.OD.nsName:
            colspan = 1
            attrSpan = self.tags['number-columns-spanned']
            if attrSpan in attrs:
                colspan = int(attrs[attrSpan])
            self.getTable().curColIndex += colspan
        elif tag == self.tags['table-column']:
            table = self.getTable()
            cols = self.tags['number-columns-repeated']
            if cols in attrs:
                table.nbOfColumns += int(attrs[cols])
            else:
                table.nbOfColumns += 1
        return ns

    def onEndElement(self):
        ns = self.namespaces
        if self.currentTag.nsName == Table.OD.nsName:
            self.tableStack.pop()
            self.tableIndex -= 1
        return ns

    def addSubBuffer(self):
        subBuffer = self.currentBuffer.addSubBuffer()
        self.currentBuffer = subBuffer
        self.mode = self.ADD_IN_BUFFER

    def propagateNamespaces(self):
        '''Propagates the namespaces in all XML element definitions that are
           used throughout POD.'''
        ns = self.namespaces
        for elemName in PodElement.POD_ELEMS:
            xmlElemDef = eval(elemName[0].upper() + elemName[1:]).OD
            elemFullName = xmlElemDef.getFullName(ns)
            xmlElemDef.__init__(elemFullName)
        # Create a table of names of used tags and attributes (precomputed,
        # including namespace, for performance).
        table = ns[self.NS_TABLE]
        text = ns[self.NS_TEXT]
        office = ns[self.NS_OFFICE]
        tags = {
          'tracked-changes': '%s:tracked-changes' % text,
          'change': '%s:change' % text,
          'annotation': '%s:annotation' % office,
          'table': '%s:table' % table,
          'table-name': '%s:name' % table,
          'table-cell': '%s:table-cell' % table,
          'table-column': '%s:table-column' % table,
          'formula': '%s:formula' % table,
          'value-type': '%s:value-type' % office,
          'value': '%s:value' % office,
          'string-value': '%s:string-value' % office,
          'span': '%s:span' % text,
          'number-columns-spanned': '%s:number-columns-spanned' % table,
          'number-columns-repeated': '%s:number-columns-repeated' % table,
        }
        self.tags = tags
        self.ignorableTags = (tags['tracked-changes'], tags['change'])
        self.exprStartElems = [self.exprStartTags[holder] \
                               for holder in self.expressionsHolders]
        self.exprEndElems = [self.exprEndTags[holder] \
                             for holder in self.expressionsHolders]
        self.impactableTags = (Text.OD.nsName, Title.OD.nsName, Item.OD.nsName,
                               Table.OD.nsName, Row.OD.nsName, Cell.OD.nsName,
                               Section.OD.nsName, Frame.OD.nsName)
        self.inserts = self.transformInserts()

    def getExpression(self, elem):
        '''We have found a pod expression within p_elem, get its value'''
        # Expression may have been found at various places
        if self.currentDbExpression:
            r = self.currentDbExpression
            self.currentDbExpression = None
        else:
            r = self.currentContent
        self.currentContent = ''
        return r.strip()

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class PodParser(OdfParser):
    def __init__(self, env, caller):
        OdfParser.__init__(self, env, caller)
        env.raiseOnError = caller.raiseOnError

    def endDocument(self):
        self.env.currentBuffer.content.close()

    def startElement(self, tag, attrs):
        e = OdfParser.startElement(self, tag, attrs)
        ns = e.onStartElement(tag, attrs)
        officeNs = ns[e.NS_OFFICE]
        textNs = ns[e.NS_TEXT]
        tableNs = ns[e.NS_TABLE]
        if tag in e.ignorableTags:
            e.state = e.IGNORING
        elif tag == e.tags['annotation']:
            # Be it in an ODT or ODS template, an annotation is considered to
            # contain a POD statement.
            e.state = e.READING_STATEMENT
        elif tag in e.exprStartElems:
            # Any track-changed text or being in a conditional or input field is
            # considered to be a POD expression.
            e.state = e.READING_EXPRESSION
            e.exprHasStyle = False
            if (tag == 'text:database-display') and \
               ('text:column-name' in attrs):
                e.currentDbExpression = attrs['text:column-name']
        elif (tag == e.tags['table-cell']) and e.tags['formula'] in attrs and \
             e.tags['value-type'] in attrs and \
             (attrs[e.tags['value-type']] == 'string') and \
             attrs[e.tags['formula']].startswith('of:="'):
            # In an ODS template, any cell containing a formula of type "string"
            # and whose content is expressed as a string between double quotes
            # (="...") is considered to contain a POD expression. But here it
            # is a special case: we need to dump the cell; the expression is not
            # directly contained within this cell; the expression will be
            # contained in the next inner paragraph. So we must here dump the
            # cell, but without some attributes, because the "formula" will be
            # converted to the result of evaluating the POD expression.
            if e.mode == e.ADD_IN_SUBBUFFER:
                e.addSubBuffer()
            e.currentBuffer.addElement(e.currentTag.name)
            hook = e.currentBuffer.dumpStartElement(tag, attrs,
                     ignoreAttrs=(e.tags['formula'], e.tags['string-value'],
                                  e.tags['value-type']),
                     hook=True)
            # We already have the POD expression: remember it on the env
            e.currentOdsExpression = attrs[e.tags['string-value']]
            e.currentOdsHook = hook
        else:
            if e.state == e.IGNORING:
                pass
            elif e.state == e.READING_CONTENT:
                if tag in e.impactableTags:
                    if e.mode == e.ADD_IN_SUBBUFFER:
                        e.addSubBuffer()
                    e.currentBuffer.addElement(e.currentTag.name)
                e.currentBuffer.dumpStartElement(tag, attrs)
            elif e.state == e.READING_STATEMENT:
                pass
            elif e.state == e.READING_EXPRESSION:
                if (tag == (e.tags['span'])) and not e.currentContent.strip():
                    e.currentBuffer.dumpStartElement(tag, attrs)
                    e.exprHasStyle = True
        e.manageInserts(tag)

    def endElement(self, tag):
        e = self.env
        ns = e.onEndElement()
        current = e.currentTag
        OdfParser.endElement(self, tag) # Pops the currently walked element
        officeNs = ns[e.NS_OFFICE]
        textNs = ns[e.NS_TEXT]
        if tag in e.ignorableTags:
            e.state = e.READING_CONTENT
        elif tag == e.tags['annotation']:
            # Manage statement
            oldCb = e.currentBuffer
            actionElemIndex = oldCb.createPodActions(e.currentStatement)
            e.currentStatement = []
            if actionElemIndex != -1:
                e.currentBuffer = oldCb.\
                    transferActionIndependentContent(actionElemIndex)
                if e.currentBuffer == oldCb:
                    e.mode = e.ADD_IN_SUBBUFFER
                else:
                    e.mode = e.ADD_IN_BUFFER
            e.state = e.READING_CONTENT
        else:
            if e.state == e.IGNORING:
                pass
            elif e.state == e.READING_CONTENT:
                # Dump the ODS POD expression if any
                if e.currentOdsExpression:
                    e.currentBuffer.addExpression(e.currentOdsExpression,
                                                  tiedHook=e.currentOdsHook)
                    e.currentOdsExpression = None
                    e.currentOdsHook = None
                # Dump the ending tag
                e.currentBuffer.dumpEndElement(tag)
                if tag in e.impactableTags:
                    if isinstance(e.currentBuffer, MemoryBuffer):
                        isMainElement = e.currentBuffer.isMainElement(tag)
                        # Unreference the element among buffer.elements
                        e.currentBuffer.unreferenceElement(tag)
                        if isMainElement:
                            parent = e.currentBuffer.parent
                            if not e.currentBuffer.action:
                                # Delete this buffer and transfer content to
                                # parent.
                                e.currentBuffer.transferAllContent()
                                parent.removeLastSubBuffer()
                                e.currentBuffer = parent
                            else:
                                if isinstance(parent, FileBuffer):
                                    # Execute buffer action and delete the
                                    # buffer.
                                    e.currentBuffer.action.execute(parent,
                                                                   e.context)
                                    parent.removeLastSubBuffer()
                                e.currentBuffer = parent
                            e.mode = e.ADD_IN_SUBBUFFER
            elif e.state == e.READING_STATEMENT:
                if tag == Text.OD.nsName:
                    statementLine = e.currentContent.strip()
                    if statementLine:
                        e.currentStatement.append(statementLine)
                    e.currentContent = ''
            elif e.state == e.READING_EXPRESSION:
                if tag in e.exprEndElems:
                    expression = e.getExpression(tag)
                    # Manage expression
                    e.currentBuffer.addExpression(expression, current)
                    if e.exprHasStyle:
                        e.currentBuffer.dumpEndElement(e.tags['span'])
                    e.state = e.READING_CONTENT

    def characters(self, content):
        e = OdfParser.characters(self, content)
        if e.state == e.IGNORING:
            pass
        elif e.state == e.READING_CONTENT:
            if e.currentOdsExpression:
                # Do not write content if we have encountered an ODS expression:
                # we will replace this content with the expression's result.
                pass
            else:
                e.currentBuffer.dumpContent(content)
        elif e.state == e.READING_STATEMENT:
            # Ignore note meta-data: creator, date, sender-initials.
            if e.currentTag.nsName in e.NOTE_TAGS:
                e.currentContent += content
        elif e.state == e.READING_EXPRESSION:
            e.currentContent += content
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''Base template for any UI page'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.px import Px

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Template:
    '''The main UI template'''

    @classmethod
    def getPageTitle(class_, tool, o):
        '''Returns content for tag html > head > title'''
        # Return the app name in some cases
        if not o or (o == tool):
            r = tool.translate('app_name_raw')
        else:
            # The page title is based on p_o's title
            r = o.getShownValue()
        # Ensure this is a string
        r = r if isinstance(r, str) else str(r)
        # If a XHTML tag is found, keep the part before it
        if '<' in r: r = r[:r.index('<')]
        # Add p_o's page title if found
        if o and ('page' in tool.req):
            label = '%s_page_%s' % (o.class_.name, tool.req.page)
            text = tool.translate(label, blankOnError=True)
            if text:
                if r: r += ' :: '
                r += text
        return Px.truncateValue(r, width=100)

    @classmethod
    def getContentCss(class_, ctx):
        '''Return the CSS class to apply to the zone of the template that will
           host the effective page content.'''
        # In most situations, class "content" must be used. In situations where
        # a more compact design must be applied (ie, in the popup), class
        # "contentP" must be used instead.
        if ctx.popup or (not ctx.showPortlet and \
                         (ctx._px_.name in ('public', 'homes'))):
            r = 'contentP'
        else:
            r = 'content'
        return r

    # PX for which compact CSS must be used
    compactPx = {'payload': ('homes',), 'content': ('public', 'homes')}

    # CSS classes to use, for every zone and compactness
    zoneCss = {'payload': {False: 'payload', True: 'payload payloadP'},
               'content': {False: 'content', True: 'contentP'}}

    @classmethod
    def getCssFor(class_, zone, ctx):
        '''Return the CSS class(es) to apply to this p_zone'''
        # In some situations (ie, in the popup), a more compact design must be
        # applied.
        compact = ctx.popup or (not ctx.showPortlet and \
                                (ctx._px_.name in (class_.compactPx[zone])))
        return class_.zoneCss[zone][compact]

    @classmethod
    def includeLoginBox(class_, ctx):
        '''Must the login box be integrated in the current page ?'''
        # The login box has no sense if the user is already authenticated or in
        # the process of self-registering himself, or if browser incompatibility
        # has be detected.
        if not ctx.isAnon or ctx.o.isTemp() or ctx.bi: return
        # Moreover, if we are on the public page and the authentication will be
        # done by redirecting the user to tool/home, it is useless to include
        # the login box in this case, too.
        if (ctx._px_.name == 'public') and (ctx.cfg.discreetLogin == 'home'):
            return
        # Include it in any other case
        return True

    @classmethod
    def getMainStyle(class_, ctx):
        '''Return the CSS properties for the main, global zone'''
        # Get the main background image
        r = ctx.cfg.getBackground(ctx._px_, ctx, type='home')
        # Get the global height. Deduce the footer height if the footer is shown
        if ctx.showFooter:
            height = 'calc(100%% - %s)' % ctx.cfg.footerHeight
        else:
            height = '100%'
        height = 'height:%s' % height
        return '%s;%s' % (r, height) if r else height

    @classmethod
    def getPayloadStyle(class_, ctx):
        '''Gets the CSS styles to apply to the "payload" zone'''
        # The payload zone is the block portlet / content / sidebar
        # ~~~ Determine the background image
        r = ctx.cfg.getBackground(ctx._px_, ctx,
                                  type='popup' if ctx.popup else 'base')
        # ~~~ Determine the "top" property
        showHeader = ctx.showHeader
        if ctx.showHeader is None or ctx.showHeader == 'sub' or \
           ctx.payloadCss.endswith('P'):
            top = '0'
        else:
            top = ctx.headerHeight
        top = 'top:%s' % top
        r = '%s;%s' % (r, top) if r else top
        return r

    @classmethod
    def getIconsStyle(class_, ctx):
        '''Get CSS styles to apply to the block of icons in the page header'''
        margin = ctx.cfg.cget('headerMargin', ctx)
        return 'margin-%s:auto' % margin

    @classmethod
    def getContent(class_, ctx):
        '''Returns the page content, with the sub-header when relevant'''
        r = '<div class="%s">%s</div>' % (class_.getCssFor('content', ctx),
                                          ctx.content(ctx, applyTemplate=False))
        # Add the sub-header when appropriate
        if ctx.showHeader == 'sub':
            r = '<div class="subbed">%s%s</div>' % (class_.pxHeader(ctx), r)
        return r

    @classmethod
    def showConnect(class_, ctx):
        '''Show the "connect" link for anons only, if discreet login is enabled,
           and if we are not already on a page allowing to log in.'''
        return ctx.isAnon and ctx.cfg.discreetLogin and \
               not ctx._px_.isHome(ctx, orPublic=False)

    # Hooks for defining PXs proposing additional links or icons, before and
    # after the links / icons corresponding to top-level pages and icons.
    pxLinksBefore = pxLinks = pxLinksAfter = None

    # Global links, that can be shown within the template header but also from
    # other pages, like "home[s]", without the base header container.
    pxTemplateLinks = Px('''
     <!-- Custom links (I) -->
     <x>:Template.pxLinksBefore</x>

     <!-- The burger button for collapsing the portlet -->
     <a if="showPortlet" class="burger"
        onclick="toggleCookie('appyPortlet','block','expanded',\
              'show','hide')"><img src=":url('burger.svg')" class="icon"/></a>

     <!-- Links and icons -->
     <div class="topIcons" style=":Template.getIconsStyle(_ctx_)">

      <!-- Custom links (II) -->
      <x>:Template.pxLinks</x>

      <!-- Header messages -->
      <span class="topText" var="text=cfg.getHeaderText(tool)"
            if="not popup and text">::text</span>

      <!-- The home icon -->
      <a if="not isAnon and cfg.tget('showHomeIcon', tool)"
         href=":tool.computeHomePage()">
        <img src=":url('home.svg')" class="icon"/></a>

      <!-- Connect link if discreet login -->
      <a if="Template.showConnect(_ctx_)" id="loginIcon" name="loginIcon"
         href=":guard.getLoginUrl(_ctx_)" class="clickable">
       <img if="cfg.showLoginIcon" src=":url('login.svg')" class="icon"/>
       <span class="topText">:_('app_connect')</span>
      </a>

      <!-- Root pages -->
      <x if="cfg.tget('showRootPages', tool)"
         var2="pages=tool.OPage.getRoot(tool)">
       <x if="pages">:tool.OPage.pxSelector</x></x>

      <!-- Language selector -->
      <x if="ui.Language.showSelector(cfg,layout)">:ui.Language.pxSelector</x>

      <!-- User info and controls for authenticated users -->
      <x if="not isAnon" var2="logoutText=_('app_logout')">
       <!-- Config -->
       <a if="cfg.showTool(tool)" href=":'%s/view' % tool.url"
              title=":_('Tool')">
        <img src=":url('config.svg')" class="icon"/></a>
       <x if="cfg.tget('showUserLink', tool)">:user.pxUserLink</x>
       <!-- Authentication context selector -->
       <x var="ctx=config.security.authContext" if="ctx">:ctx.pxLogged</x>
       <!-- Log out -->
       <a href=":guard.getLogoutUrl(tool, user)" title=":logoutText">
        <img src=":url('logout.svg')" class="icon"/>
        <span if="cfg.logoutText">:logoutText</span></a>
      </x>

      <!-- Custom links (III) -->
      <x>:Template.pxLinksAfter</x>
     </div>

     <!-- The burger button for collapsing the sidebar -->
     <a if="showSidebar" class="burger"
        onclick="toggleCookie('appySidebar','block','expanded',\
           'show','hide')"><img src=":url('burger.svg')" class="icon"/></a>
     ''',

     css='''
      .topIcons > a, .topIcons > select { margin:0 5px; color:|headerColor| }
      .topIcons span { text-transform: uppercase; font-weight: bold;
                       padding: 0 15px; vertical-align: -2px }
      .topText { letter-spacing:1px; padding: 0 8px }
      .burger { cursor:pointer; margin:0 5px }''')

    # The global page header
    pxHeader = Px('''
     <div class="topBase"
          style=":cfg.getBackground(_px_, _ctx_, type='header')">
      <div class="top" style=":'height:%s' % headerHeight">
       <x>:Template.pxTemplateLinks</x>
      </div>
     </div>''',

     css='''
      .topBase { background-color:|headerBgColor|; width:100%; position:fixed;
                 color:|headerColor|; font-weight:lighter; z-index:1 }
      .top { display:flex; flex-wrap:nowrap; justify-content:space-between;
             align-items:center; margin:0 5px }
     ''')

    # The template of all base PXs
    px = Px('''
     <html var="x=handler.customInit(); cfg=config.ui; Template=ui.Template"
           dir=":dir">

      <head>
       <title>:Template.getPageTitle(tool, o or home)</title>
       <meta http-equiv="X-UA-Compatible" content="IE=edge,chrome=1"/>
       <link rel="icon" type="image/x-icon"
             href=":url('favicon.ico', base=appName)"/>
       <link rel="apple-touch-icon" href=":url('appleicon', base=appName)"/>
       <x>::ui.Includer.getGlobal(handler, config, dir)</x>
      </head>

      <body class=":cfg.getClass('body', _px_, _ctx_)"
            var="showPortlet=ui.Portlet.show(tool, _px_, _ctx_);
                 showSidebar=ui.Sidebar.show(tool, o, layout, popup);
                 showHeader=cfg.showHeader(_px_, _ctx_, popup);
                 headerHeight=cfg.cget('headerHeight', _ctx_);
                 showFooter=cfg.showFooter(_px_, _ctx_, popup);
                 bi=ui.Browser.getIncompatibilityMessage(tool, handler)">

       <!-- Browser incompatibility message -->
       <div if="bi" class="wrongBrowser">::bi</div>

       <!-- Google Analytics stuff, if enabled -->
       <script var="gaCode=tool.getGoogleAnalyticsCode(handler, config)"
               if="gaCode">:gaCode</script>

       <!-- Popups -->
       <x>::ui.Globals.getPopups(tool, url, _, dleft, dright, popup)</x>

       <div class=":cfg.getClass('main', _px_, _ctx_)"
            style=":Template.getMainStyle(_ctx_)">

        <!-- Page header (top) -->
        <x if="showHeader == 'top'">:Template.pxHeader</x>

        <!-- Message zone -->
        <div height="0">:ui.Message.px</div>

        <!-- Login zone -->
        <x if="Template.includeLoginBox(_ctx_)">:guard.pxLogin</x>

        <!-- Payload: portlet / content / sidebar -->
        <div var="payloadCss=Template.getCssFor('payload', _ctx_)"
             class=":payloadCss" style=":Template.getPayloadStyle(_ctx_)">

         <!-- Portlet -->
         <x if="showPortlet">:ui.Portlet.px</x>

         <!-- Page content -->
         <x>::Template.getContent(_ctx_)</x>

         <!-- Sidebar -->
         <x if="showSidebar">:ui.Sidebar.px</x>
        </div>

        <!-- Cookie warning -->
        <x if="not popup and not guard.cookiesAccepted">:guard.pxWarnCookies</x>

        <!-- Footer -->
        <x if="showFooter">:ui.Footer.px</x>

       </div>
      </body>
     </html>''', prologue=Px.xhtmlPrologue)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
